from __future__ import annotations

from collections.abc import Callable

import casadi as ca

from ..mtor_types import Constraint, PhaseID
from .state import MultiPhaseVariableState, PhaseDefinition, _BoundaryConstraint


def _add_path_constraint(phase_def: PhaseDefinition, constraint_expr: ca.MX | float | int) -> None:
    if isinstance(constraint_expr, ca.MX):
        phase_def.path_constraints.append(constraint_expr)
    else:
        phase_def.path_constraints.append(ca.MX(constraint_expr))


def _add_event_constraint(
    multiphase_state: MultiPhaseVariableState, constraint_expr: ca.MX | float | int
) -> None:
    if isinstance(constraint_expr, ca.MX):
        multiphase_state.cross_phase_constraints.append(constraint_expr)
    else:
        multiphase_state.cross_phase_constraints.append(ca.MX(constraint_expr))


def _symbolic_constraint_to_constraint(expr: ca.MX) -> Constraint:
    try:
        OP_EQ = getattr(ca, "OP_EQ", None)
        OP_LE = getattr(ca, "OP_LE", None)
        OP_GE = getattr(ca, "OP_GE", None)

        if (
            isinstance(expr, ca.MX)
            and hasattr(expr, "is_op")
            and OP_EQ is not None
            and isinstance(OP_EQ, int)
        ):
            if expr.is_op(OP_EQ):
                lhs = expr.dep(0)
                rhs = expr.dep(1)
                return Constraint(val=lhs - rhs, equals=0.0)

            elif OP_LE is not None and isinstance(OP_LE, int) and expr.is_op(OP_LE):
                lhs = expr.dep(0)
                rhs = expr.dep(1)
                return Constraint(val=lhs - rhs, max_val=0.0)

            elif OP_GE is not None and isinstance(OP_GE, int) and expr.is_op(OP_GE):
                lhs = expr.dep(0)
                rhs = expr.dep(1)
                return Constraint(val=lhs - rhs, min_val=0.0)

    except (AttributeError, TypeError, NotImplementedError):
        pass

    return Constraint(val=expr, equals=0.0)


def _boundary_constraint_to_constraints(
    boundary_constraint: _BoundaryConstraint, variable_expression: ca.MX
) -> list[Constraint]:
    constraints: list[Constraint] = []

    if boundary_constraint.equals is not None:
        constraints.append(Constraint(val=variable_expression, equals=boundary_constraint.equals))
    else:
        if boundary_constraint.lower is not None:
            constraints.append(
                Constraint(val=variable_expression, min_val=boundary_constraint.lower)
            )
        if boundary_constraint.upper is not None:
            constraints.append(
                Constraint(val=variable_expression, max_val=boundary_constraint.upper)
            )

    return constraints


def _map_symbol_to_vector_element(
    symbol: ca.MX, vector: ca.MX, index: int, vector_length: int
) -> ca.MX:
    return vector if vector_length == 1 else vector[index]


def _process_boundary_constraints(
    boundary_constraints: list[_BoundaryConstraint | None],
    variables_vec: ca.MX,
    result: list[Constraint],
) -> None:
    for i, boundary_constraint in enumerate(boundary_constraints):
        if boundary_constraint is None or not boundary_constraint.has_constraint():
            continue

        variable_expr = _map_symbol_to_vector_element(
            variables_vec, variables_vec, i, len(boundary_constraints)
        )
        result.extend(_boundary_constraint_to_constraints(boundary_constraint, variable_expr))


def _process_state_boundary_constraints(
    state_boundary_constraints: list[_BoundaryConstraint | None],
    states_vec: ca.MX,
    result: list[Constraint],
) -> None:
    _process_boundary_constraints(state_boundary_constraints, states_vec, result)


def _process_control_boundary_constraints(
    control_boundary_constraints: list[_BoundaryConstraint | None],
    controls_vec: ca.MX,
    result: list[Constraint],
) -> None:
    _process_boundary_constraints(control_boundary_constraints, controls_vec, result)


def _process_symbolic_path_constraints(
    path_constraints: list[ca.MX],
    subs_map: dict[ca.MX, ca.MX],
    result: list[Constraint],
) -> None:
    for expr in path_constraints:
        substituted_expr = ca.substitute([expr], list(subs_map.keys()), list(subs_map.values()))[0]
        result.append(_symbolic_constraint_to_constraint(substituted_expr))


def _map_phase_symbols_to_substitution(
    phase_def: PhaseDefinition,
    states_vec: ca.MX,
    controls_vec: ca.MX,
    time: ca.MX,
    initial_time_variable: ca.MX | None,
    terminal_time_variable: ca.MX | None,
) -> dict[ca.MX, ca.MX]:
    subs_map = {}

    state_syms = phase_def._get_ordered_state_symbols()
    for i, state_sym in enumerate(state_syms):
        subs_map[state_sym] = _map_symbol_to_vector_element(
            state_sym, states_vec, i, len(state_syms)
        )

    control_syms = phase_def._get_ordered_control_symbols()
    for i, control_sym in enumerate(control_syms):
        subs_map[control_sym] = _map_symbol_to_vector_element(
            control_sym, controls_vec, i, len(control_syms)
        )

    if phase_def.sym_time is not None:
        subs_map[phase_def.sym_time] = time

    if phase_def.sym_time_initial is not None and initial_time_variable is not None:
        subs_map[phase_def.sym_time_initial] = initial_time_variable

    if phase_def.sym_time_final is not None and terminal_time_variable is not None:
        subs_map[phase_def.sym_time_final] = terminal_time_variable

    return subs_map


def _map_static_parameters_to_substitution(
    static_parameters_vec: ca.MX | None,
    static_parameter_symbols: list[ca.MX] | None,
    subs_map: dict[ca.MX, ca.MX],
) -> None:
    if static_parameters_vec is not None and static_parameter_symbols is not None:
        num_params = len(static_parameter_symbols)
        for i, param_sym in enumerate(static_parameter_symbols):
            subs_map[param_sym] = _map_symbol_to_vector_element(
                param_sym, static_parameters_vec, i, num_params
            )


def _build_substitution_map(
    phase_def: PhaseDefinition,
    states_vec: ca.MX,
    controls_vec: ca.MX,
    time: ca.MX,
    static_parameters_vec: ca.MX | None,
    static_parameter_symbols: list[ca.MX] | None,
    initial_time_variable: ca.MX | None,
    terminal_time_variable: ca.MX | None,
) -> dict[ca.MX, ca.MX]:
    subs_map = _map_phase_symbols_to_substitution(
        phase_def, states_vec, controls_vec, time, initial_time_variable, terminal_time_variable
    )

    _map_static_parameters_to_substitution(
        static_parameters_vec, static_parameter_symbols, subs_map
    )

    return subs_map


def _check_has_path_constraints(phase_def: PhaseDefinition) -> bool:
    has_path_constraints = bool(phase_def.path_constraints)

    state_boundary_constraints = [info.boundary_constraint for info in phase_def.state_info]
    control_boundary_constraints = [info.boundary_constraint for info in phase_def.control_info]

    has_state_boundary = any(
        constraint is not None and constraint.has_constraint()
        for constraint in state_boundary_constraints
    )
    has_control_boundary = any(
        constraint is not None and constraint.has_constraint()
        for constraint in control_boundary_constraints
    )

    return has_path_constraints or has_state_boundary or has_control_boundary


def _create_path_constraints(
    phase_def: PhaseDefinition,
) -> Callable[..., list[Constraint]]:
    state_boundary_constraints = [info.boundary_constraint for info in phase_def.state_info]
    control_boundary_constraints = [info.boundary_constraint for info in phase_def.control_info]

    def _path_constraints(
        states_vec: ca.MX,
        controls_vec: ca.MX,
        time: ca.MX,
        static_parameters_vec: ca.MX | None = None,
        static_parameter_symbols: list[ca.MX] | None = None,
        initial_time_variable: ca.MX | None = None,
        terminal_time_variable: ca.MX | None = None,
    ) -> list[Constraint]:
        result: list[Constraint] = []

        subs_map = _build_substitution_map(
            phase_def,
            states_vec,
            controls_vec,
            time,
            static_parameters_vec,
            static_parameter_symbols,
            initial_time_variable,
            terminal_time_variable,
        )

        _process_symbolic_path_constraints(phase_def.path_constraints, subs_map, result)
        _process_state_boundary_constraints(state_boundary_constraints, states_vec, result)
        _process_control_boundary_constraints(control_boundary_constraints, controls_vec, result)

        return result

    return _path_constraints


def _get_phase_path_constraints_function(
    phase_def: PhaseDefinition,
) -> Callable[..., list[Constraint]] | None:
    if not _check_has_path_constraints(phase_def):
        return None

    return _create_path_constraints(phase_def)


def _map_phase_time_symbols(
    phase_def: PhaseDefinition, endpoint_data: dict[str, ca.MX], subs_map: dict[ca.MX, ca.MX]
) -> None:
    if phase_def.sym_time_initial is not None:
        subs_map[phase_def.sym_time_initial] = endpoint_data["t0"]
    if phase_def.sym_time_final is not None:
        subs_map[phase_def.sym_time_final] = endpoint_data["tf"]
    if phase_def.sym_time is not None:
        subs_map[phase_def.sym_time] = endpoint_data["tf"]


def _map_phase_state_symbols(
    phase_def: PhaseDefinition, endpoint_data: dict[str, ca.MX], subs_map: dict[ca.MX, ca.MX]
) -> None:
    """Map phase state symbols with guaranteed non-None symbols."""
    state_initial_syms = phase_def._get_ordered_state_initial_symbols()  # Now returns list[ca.MX]
    state_final_syms = phase_def._get_ordered_state_final_symbols()  # Now returns list[ca.MX]
    state_syms = phase_def._get_ordered_state_symbols()

    x0_vec = endpoint_data["x0"]
    xf_vec = endpoint_data["xf"]
    num_states = len(state_syms)

    for i, (sym_initial, sym_final, sym_current) in enumerate(
        zip(state_initial_syms, state_final_syms, state_syms, strict=True)
    ):
        subs_map[sym_initial] = _map_symbol_to_vector_element(sym_initial, x0_vec, i, num_states)
        subs_map[sym_final] = _map_symbol_to_vector_element(sym_final, xf_vec, i, num_states)
        subs_map[sym_current] = _map_symbol_to_vector_element(sym_current, xf_vec, i, num_states)


def _map_phase_integral_symbols(
    phase_def: PhaseDefinition, endpoint_data: dict[str, ca.MX], subs_map: dict[ca.MX, ca.MX]
) -> None:
    if "q" in endpoint_data and endpoint_data["q"] is not None:
        for i, integral_sym in enumerate(phase_def.integral_symbols):
            if i < endpoint_data["q"].shape[0]:
                subs_map[integral_sym] = endpoint_data["q"][i]


def _map_cross_phase_static_parameters(
    multiphase_state: MultiPhaseVariableState,
    static_parameters_vec: ca.MX | None,
    subs_map: dict[ca.MX, ca.MX],
) -> None:
    if static_parameters_vec is not None:
        static_param_syms = multiphase_state.static_parameters.get_ordered_parameter_symbols()
        num_params = len(static_param_syms)
        for i, param_sym in enumerate(static_param_syms):
            subs_map[param_sym] = _map_symbol_to_vector_element(
                param_sym, static_parameters_vec, i, num_params
            )


def _process_cross_phase_substitution_map(
    multiphase_state: MultiPhaseVariableState,
    phase_endpoint_vectors: dict[PhaseID, dict[str, ca.MX]],
    static_parameters_vec: ca.MX | None,
) -> dict[ca.MX, ca.MX]:
    """Process cross-phase substitution mapping with proper type annotation."""
    subs_map: dict[ca.MX, ca.MX] = {}

    for phase_id, phase_def in multiphase_state.phases.items():
        if phase_id not in phase_endpoint_vectors:
            continue

        endpoint_data = phase_endpoint_vectors[phase_id]

        _map_phase_time_symbols(phase_def, endpoint_data, subs_map)
        _map_phase_state_symbols(phase_def, endpoint_data, subs_map)
        _map_phase_integral_symbols(phase_def, endpoint_data, subs_map)

    _map_cross_phase_static_parameters(multiphase_state, static_parameters_vec, subs_map)

    return subs_map


def _process_cross_phase_symbolic_constraints(
    multiphase_state: MultiPhaseVariableState,
    subs_map: dict[ca.MX, ca.MX],
    result: list[Constraint],
) -> None:
    for expr in multiphase_state.cross_phase_constraints:
        substituted_expr = ca.substitute([expr], list(subs_map.keys()), list(subs_map.values()))[0]
        constraint = _symbolic_constraint_to_constraint(substituted_expr)
        result.append(constraint)


def _process_phase_endpoint_boundary_constraints(
    multiphase_state: MultiPhaseVariableState,
    phase_endpoint_vectors: dict[PhaseID, dict[str, ca.MX]],
    constraint_type: str,
    result: list[Constraint],
) -> None:
    for phase_id, phase_def in multiphase_state.phases.items():
        if phase_id not in phase_endpoint_vectors:
            continue

        endpoint_data = phase_endpoint_vectors[phase_id]

        if constraint_type == "initial":
            boundary_constraints = [info.initial_constraint for info in phase_def.state_info]
            endpoint_vec = endpoint_data["x0"]
        else:
            boundary_constraints = [info.final_constraint for info in phase_def.state_info]
            endpoint_vec = endpoint_data["xf"]

        num_states = len(phase_def.state_info)
        for i, boundary_constraint in enumerate(boundary_constraints):
            if (
                boundary_constraint is None
                or not boundary_constraint.has_constraint()
                or boundary_constraint.is_symbolic()
            ):
                continue

            constraint_expr = _map_symbol_to_vector_element(
                endpoint_vec, endpoint_vec, i, num_states
            )
            result.extend(_boundary_constraint_to_constraints(boundary_constraint, constraint_expr))


def _process_phase_initial_boundary_constraints(
    multiphase_state: MultiPhaseVariableState,
    phase_endpoint_vectors: dict[PhaseID, dict[str, ca.MX]],
    result: list[Constraint],
) -> None:
    _process_phase_endpoint_boundary_constraints(
        multiphase_state, phase_endpoint_vectors, "initial", result
    )


def _process_phase_final_boundary_constraints(
    multiphase_state: MultiPhaseVariableState,
    phase_endpoint_vectors: dict[PhaseID, dict[str, ca.MX]],
    result: list[Constraint],
) -> None:
    _process_phase_endpoint_boundary_constraints(
        multiphase_state, phase_endpoint_vectors, "final", result
    )


def _check_has_cross_phase_constraints(multiphase_state: MultiPhaseVariableState) -> bool:
    has_cross_phase_constraints = bool(multiphase_state.cross_phase_constraints)

    has_phase_event_constraints = False
    for phase_def in multiphase_state.phases.values():
        state_initial_constraints = [info.initial_constraint for info in phase_def.state_info]
        state_final_constraints = [info.final_constraint for info in phase_def.state_info]

        if any(
            constraint is not None and constraint.has_constraint() and not constraint.is_symbolic()
            for constraint in (state_initial_constraints + state_final_constraints)
        ):
            has_phase_event_constraints = True
            break

    return has_cross_phase_constraints or has_phase_event_constraints


def _create_cross_phase_event_constraints(
    multiphase_state: MultiPhaseVariableState,
) -> Callable[..., list[Constraint]]:
    def _cross_phase_event_constraints(
        phase_endpoint_vectors: dict[PhaseID, dict[str, ca.MX]], static_parameters_vec: ca.MX | None
    ) -> list[Constraint]:
        result: list[Constraint] = []

        subs_map = _process_cross_phase_substitution_map(
            multiphase_state, phase_endpoint_vectors, static_parameters_vec
        )

        _process_cross_phase_symbolic_constraints(multiphase_state, subs_map, result)
        _process_phase_initial_boundary_constraints(
            multiphase_state, phase_endpoint_vectors, result
        )
        _process_phase_final_boundary_constraints(multiphase_state, phase_endpoint_vectors, result)

        return result

    return _cross_phase_event_constraints


def _get_cross_phase_event_constraints_function(
    multiphase_state: MultiPhaseVariableState,
) -> Callable[..., list[Constraint]] | None:
    if not _check_has_cross_phase_constraints(multiphase_state):
        return None

    return _create_cross_phase_event_constraints(multiphase_state)



import casadi as ca
import numpy as np
import pytest

import maptor as mtor
from maptor.exceptions import ConfigurationError


class TestProvenExamples:
    def test_simple_lqr_problem_solution(self):
        problem = mtor.Problem("LQR Test")
        phase = problem.set_phase(1)

        _t = phase.time(initial=0, final=1)
        x = phase.state("x", initial=1.0)
        u = phase.control("u")

        phase.dynamics({x: 0.5 * x + u})

        integrand = 0.625 * x**2 + 0.5 * x * u + 0.5 * u**2
        integral_var = phase.add_integral(integrand)
        problem.minimize(integral_var)

        phase.mesh([4, 4], [-1.0, 0.0, 1.0])

        # Simple initial guess matching examples/lqr pattern
        states_guess = [
            np.array([[1.0, 0.8, 0.6, 0.4, 0.2]]),
            np.array([[0.2, 0.15, 0.1, 0.05, 0.0]]),
        ]
        controls_guess = [
            np.array([[-0.5, -0.4, -0.3, -0.2]]),
            np.array([[-0.1, -0.05, 0.0, 0.05]]),
        ]

        phase.guess(
            states=states_guess,
            controls=controls_guess,
            integrals=0.5,
        )

        # Use public API
        solution = mtor.solve_fixed_mesh(
            problem, nlp_options={"ipopt.print_level": 0, "ipopt.max_iter": 200}
        )

        # Verify solution interface
        assert hasattr(solution, "status")
        assert isinstance(solution.status["success"], bool)
        assert isinstance(solution.status["message"], str)

        if solution.status["success"]:
            # LQR reference objective: ~0.380797077977481140
            objective = solution.status["objective"]
            assert 0.35 < objective < 0.45  # Reasonable range for LQR

            # Test data access through new interface
            x_trajectory = solution["x"]
            assert len(x_trajectory) > 0
            assert abs(x_trajectory[0] - 1.0) < 0.1  # Initial condition

            # Test phase information
            assert 1 in solution.phases
            phase_data = solution.phases[1]
            assert "times" in phase_data
            assert "variables" in phase_data
            assert "x" in phase_data["variables"]["state_names"]

    def test_minimum_time_brachistochrone(self):
        problem = mtor.Problem("Brachistochrone Test")
        phase = problem.set_phase(1)

        t = phase.time(initial=0.0)
        x = phase.state("x", initial=0.0, final=1.0, boundary=(0, 10))
        y = phase.state("y", initial=0.0, boundary=(0, 10))
        v = phase.state("v", initial=0.0, boundary=(0, 10))
        u = phase.control("u", boundary=(0, np.pi / 2))

        g0 = 32.174
        phase.dynamics({x: v * ca.cos(u), y: v * ca.sin(u), v: g0 * ca.sin(u)})

        problem.minimize(t.final)

        phase.mesh([5, 5], [-1.0, 0.0, 1.0])

        # Initial guess matching examples/brachistochrone pattern
        states_guess = []
        controls_guess = []

        for N in [5, 5]:
            x_vals = np.array([0.0, 0.25, 0.5, 0.75, 1.0, 1.0])
            y_vals = np.array([0.0, 0.1, 0.3, 0.6, 1.0, 1.2])
            v_vals = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 4.5])
            states_guess.append(np.array([x_vals, y_vals, v_vals]))

            u_vals = np.ones(N) * 0.5
            controls_guess.append(np.array([u_vals]))

        phase.guess(
            states=states_guess,
            controls=controls_guess,
            terminal_time=0.4,
        )

        solution = mtor.solve_fixed_mesh(
            problem, nlp_options={"ipopt.print_level": 0, "ipopt.max_iter": 500}
        )

        assert hasattr(solution, "status")

        if solution.status["success"]:
            # Brachistochrone reference: ~0.312480130
            objective = solution.status["objective"]
            assert 0.25 < objective < 0.40  # Reasonable range

            # Verify final condition satisfaction
            x_final = solution["x"][-1]
            assert abs(x_final - 1.0) < 0.1  # Should reach x=1

    def test_hypersensitive_problem(self):
        problem = mtor.Problem("Hypersensitive Test")
        phase = problem.set_phase(1)

        _t = phase.time(initial=0, final=40)
        x = phase.state("x", initial=1.5, final=1.0)
        u = phase.control("u")

        phase.dynamics({x: -(x**3) + u})

        integrand = 0.5 * (x**2 + u**2)
        integral_var = phase.add_integral(integrand)
        problem.minimize(integral_var)

        phase.mesh([8, 8, 8], [-1.0, -1 / 3, 1 / 3, 1.0])

        # Initial guess following examples/hypersensitive pattern
        states_guess = []
        controls_guess = []
        for N in [8, 8, 8]:
            tau = np.linspace(-1, 1, N + 1)
            x_vals = 1.5 + (1.0 - 1.5) * (tau + 1) / 2
            states_guess.append(x_vals.reshape(1, -1))
            controls_guess.append(np.zeros((1, N)))

        phase.guess(
            states=states_guess,
            controls=controls_guess,
            integrals=0.1,
        )

        solution = mtor.solve_fixed_mesh(
            problem, nlp_options={"ipopt.print_level": 0, "ipopt.max_iter": 200}
        )

        assert hasattr(solution, "status")
        # Hypersensitive is challenging - may not always converge, but should not crash

    def test_multiphase_schwartz_problem(self):
        problem = mtor.Problem("Two-Phase Test")

        # Phase 1
        phase1 = problem.set_phase(1)
        phase1.time(initial=0.0, final=1.0)
        x0_1 = phase1.state("x0", initial=1.0)
        x1_1 = phase1.state("x1", initial=1.0, boundary=(-0.8, None))
        u1 = phase1.control("u", boundary=(-1.0, 1.0))

        phase1.dynamics(
            {
                x0_1: x1_1,
                x1_1: u1 - 0.1 * (1 + 2 * x0_1**2) * x1_1,
            }
        )

        # Path constraint
        elliptical_constraint = 1 - 9 * (x0_1 - 1) ** 2 - ((x1_1 - 0.4) / 0.3) ** 2
        phase1.path_constraints(elliptical_constraint <= 0)
        phase1.mesh([6, 6], [-1.0, 0.0, 1.0])

        # Phase 2
        phase2 = problem.set_phase(2)
        phase2.time(initial=1.0, final=2.9)
        x0_2 = phase2.state("x0", initial=x0_1.final)  # Symbolic continuity
        x1_2 = phase2.state("x1", initial=x1_1.final)  # Symbolic continuity
        u2 = phase2.control("u")

        phase2.dynamics(
            {
                x0_2: x1_2,
                x1_2: u2 - 0.1 * (1 + 2 * x0_2**2) * x1_2,
            }
        )
        phase2.mesh([8, 8], [-1.0, 0.0, 1.0])

        # Objective
        objective_expr = 5 * (x0_2.final**2 + x1_2.final**2)
        problem.minimize(objective_expr)

        # Simple guess
        states_p1 = []
        controls_p1 = []
        states_p2 = []
        controls_p2 = []

        for N in [6, 6]:
            tau_states = np.linspace(-1, 1, N + 1)
            t_norm_states = (tau_states + 1) / 2
            x0_vals = 1.0 + 0.2 * t_norm_states
            x1_vals = 1.0 - 0.3 * t_norm_states
            states_p1.append(np.array([x0_vals, x1_vals]))

            t_norm_controls = np.linspace(0, 1, N)
            u_vals = 0.3 * np.sin(np.pi * t_norm_controls)
            controls_p1.append(np.array([u_vals]))

        for N in [8, 8]:
            tau_states = np.linspace(-1, 1, N + 1)
            t_norm_states = (tau_states + 1) / 2
            x0_end_p1 = 1.2
            x1_end_p1 = 0.7
            x0_vals = x0_end_p1 * (1 - 0.8 * t_norm_states)
            x1_vals = x1_end_p1 * (1 - 0.9 * t_norm_states)
            states_p2.append(np.array([x0_vals, x1_vals]))

            t_norm_controls = np.linspace(0, 1, N)
            u_vals = -1.0 + 0.5 * t_norm_controls
            controls_p2.append(np.array([u_vals]))

        phase1.guess(
            states=states_p1,
            controls=controls_p1,
        )
        phase2.guess(
            states=states_p2,
            controls=controls_p2,
        )

        solution = mtor.solve_fixed_mesh(
            problem, nlp_options={"ipopt.print_level": 0, "ipopt.max_iter": 1000}
        )

        assert hasattr(solution, "status")

        if solution.status["success"]:
            # Should have data for both phases
            assert 1 in solution.phases
            assert 2 in solution.phases

            # Test multiphase data access
            x0_phase1 = solution[(1, "x0")]
            x0_phase2 = solution[(2, "x0")]
            assert len(x0_phase1) > 0
            assert len(x0_phase2) > 0

    def test_adaptive_solver_execution(self):
        """Test adaptive solver using public API."""
        problem = mtor.Problem("Adaptive Test")
        phase = problem.set_phase(1)

        t = phase.time(initial=0, final=1)
        x = phase.state("x", initial=0, final=1)
        u = phase.control("u")

        phase.dynamics({x: u})
        problem.minimize(t.final)

        phase.mesh([3, 3], [-1, 0, 1])

        solution = mtor.solve_adaptive(
            problem,
            error_tolerance=1e-3,
            max_iterations=10,
            min_polynomial_degree=3,
            max_polynomial_degree=8,
            nlp_options={"ipopt.print_level": 0, "ipopt.max_iter": 100},
        )

        assert hasattr(solution, "status")
        assert hasattr(solution, "adaptive")

        if solution.adaptive:
            adaptive_data = solution.adaptive
            assert "converged" in adaptive_data
            assert "iterations" in adaptive_data
            assert "target_tolerance" in adaptive_data
            assert isinstance(adaptive_data["converged"], bool)


class TestSolutionInterfaceStructure:
    def test_solution_status_interface(self):
        problem = mtor.Problem("Status Interface Test")
        phase = problem.set_phase(1)

        t = phase.time(initial=0, final=1)
        x = phase.state("x", initial=0, final=1)
        u = phase.control("u")

        phase.dynamics({x: u})
        problem.minimize(t.final)

        phase.mesh([3], [-1, 1])

        solution = mtor.solve_fixed_mesh(problem, nlp_options={"ipopt.print_level": 0})

        # Test status interface matches examples pattern
        status = solution.status
        assert "success" in status
        assert "message" in status
        assert "objective" in status
        assert "total_mission_time" in status

        assert isinstance(status["success"], bool)
        assert isinstance(status["message"], str)

        if status["success"]:
            assert isinstance(status["objective"], int | float)
            assert isinstance(status["total_mission_time"], int | float)

    def test_variable_access_patterns(self):
        problem = mtor.Problem("Variable Access Test")
        phase = problem.set_phase(1)

        x = phase.state("position", initial=0, final=1)
        v = phase.state("velocity", initial=0)
        T = phase.control("thrust")

        phase.dynamics({x: v, v: T})
        problem.minimize(x.final)

        phase.mesh([3], [-1, 1])

        solution = mtor.solve_fixed_mesh(problem, nlp_options={"ipopt.print_level": 0})

        if solution.status["success"]:
            # Test different access patterns from examples
            position_data = solution["position"]  # Auto-find in any phase
            velocity_data = solution[(1, "velocity")]  # Specific phase
            thrust_data = solution[(1, "thrust")]

            assert isinstance(position_data, np.ndarray)
            assert isinstance(velocity_data, np.ndarray)
            assert isinstance(thrust_data, np.ndarray)

            # Test time arrays
            time_states = solution[(1, "time_states")]
            time_controls = solution[(1, "time_controls")]

            assert isinstance(time_states, np.ndarray)
            assert isinstance(time_controls, np.ndarray)

    def test_phases_property_interface(self):
        problem = mtor.Problem("Phases Interface Test")

        # Multiphase problem
        p1 = problem.set_phase(1)
        t1 = p1.time(initial=0, final=1)
        x1 = p1.state("x", initial=0)
        u1 = p1.control("u")
        p1.dynamics({x1: u1})
        p1.mesh([3], [-1, 1])

        p2 = problem.set_phase(2)
        t2 = p2.time(initial=t1.final, final=2)
        x2 = p2.state("x", initial=x1.final)
        u2 = p2.control("u")
        p2.dynamics({x2: u2})
        p2.mesh([4], [-1, 1])

        problem.minimize(t2.final)

        solution = mtor.solve_fixed_mesh(problem, nlp_options={"ipopt.print_level": 0})

        if solution.status["success"]:
            # Test phases interface from examples
            phases = solution.phases

            for phase_data in phases.values():
                assert "times" in phase_data
                assert "variables" in phase_data
                assert "mesh" in phase_data
                assert "time_arrays" in phase_data

                times = phase_data["times"]
                assert "initial" in times
                assert "final" in times
                assert "duration" in times

                variables = phase_data["variables"]
                assert "state_names" in variables
                assert "control_names" in variables
                assert "num_states" in variables
                assert "num_controls" in variables

                mesh = phase_data["mesh"]
                assert "polynomial_degrees" in mesh
                assert "mesh_nodes" in mesh
                assert "num_intervals" in mesh

    def test_parameters_interface(self):
        problem = mtor.Problem("Parameters Interface Test")

        mass = problem.parameter("mass", boundary=(100, 1000))
        thrust_max = problem.parameter("thrust_max", boundary=(500, 2000))

        phase = problem.set_phase(1)
        v = phase.state("velocity", initial=0)
        u = phase.control("throttle", boundary=(0, 1))

        phase.dynamics({v: u * thrust_max / mass})
        problem.minimize(v.final)

        phase.mesh([3], [-1, 1])

        solution = mtor.solve_fixed_mesh(problem, nlp_options={"ipopt.print_level": 0})

        if solution.status["success"]:
            params = solution.parameters

            if params is not None:
                assert "values" in params
                assert "names" in params
                assert "count" in params

                assert isinstance(params["values"], np.ndarray)
                assert params["count"] == 2


class TestErrorHandlingAndFailures:
    def test_infeasible_problem_handling(self):
        problem = mtor.Problem("Infeasible Problem")
        phase = problem.set_phase(1)

        x = phase.state("x", initial=0, final=10)  # Large target
        u = phase.control("u", boundary=(-0.1, 0.1))  # Very limited control

        phase.dynamics({x: u})
        problem.minimize(x.final)

        phase.mesh([3], [-1, 1])

        solution = mtor.solve_fixed_mesh(
            problem, nlp_options={"ipopt.print_level": 0, "ipopt.max_iter": 50}
        )

        # Should get solution object even if failed
        assert hasattr(solution, "status")
        assert isinstance(solution.status["success"], bool)
        assert isinstance(solution.status["message"], str)

        # May fail due to infeasibility, but should not crash

    def test_malformed_problem_validation(self):
        problem = mtor.Problem("Malformed Problem")
        phase = problem.set_phase(1)

        _x = phase.state("x", initial=0)
        # Missing dynamics, mesh, and objective

        # Should be caught by solve_fixed_mesh validation
        with pytest.raises((ValueError, ConfigurationError)):
            mtor.solve_fixed_mesh(problem)

    def test_invalid_solver_options(self):
        problem = mtor.Problem("Invalid Options Test")
        phase = problem.set_phase(1)

        x = phase.state("x", initial=0, final=1)
        u = phase.control("u")

        phase.dynamics({x: u})
        problem.minimize(x.final)

        phase.mesh([3], [-1, 1])

        # Invalid solver options should be handled gracefully
        solution = mtor.solve_fixed_mesh(problem, nlp_options={"invalid_option": "invalid_value"})

        # Should still return solution object
        assert hasattr(solution, "status")

    def test_solution_contains_operator(self):
        problem = mtor.Problem("Contains Test")
        phase = problem.set_phase(1)

        x = phase.state("position", initial=0)
        u = phase.control("thrust")

        phase.dynamics({x: u})
        problem.minimize(x.final)

        phase.mesh([3], [-1, 1])

        solution = mtor.solve_fixed_mesh(problem, nlp_options={"ipopt.print_level": 0})

        if solution.status["success"]:
            # Test contains operator
            assert "position" in solution
            assert (1, "position") in solution
            assert (1, "thrust") in solution
            assert (1, "time_states") in solution
            assert "nonexistent_var" not in solution


class TestConstraintIntegrationEndToEnd:
    def test_path_constraints_with_public_api(self):
        problem = mtor.Problem("Path Constraints Test")
        phase = problem.set_phase(1)

        x = phase.state("position", initial=0, final=1)
        v = phase.state("velocity", initial=0)
        u = phase.control("thrust")

        phase.dynamics({x: v, v: u})

        # Path constraints
        phase.path_constraints(
            x >= 0,  # Stay non-negative
            v <= 5,  # Speed limit
            u >= -2,  # Thrust bounds
            u <= 2,
        )

        problem.minimize(x.final)

        phase.mesh([4], [-1, 1])

        solution = mtor.solve_fixed_mesh(problem, nlp_options={"ipopt.print_level": 0})

        if solution.status["success"]:
            # Basic constraint satisfaction check
            position_data = solution["position"]
            velocity_data = solution["velocity"]

            # Should generally satisfy constraints (within solver tolerance)
            assert np.all(position_data >= -0.1)  # Small tolerance for numerical error
            assert np.all(velocity_data <= 5.1)  # Small tolerance

    def test_event_constraints_with_public_api(self):
        problem = mtor.Problem("Event Constraints Test")
        phase = problem.set_phase(1)

        x = phase.state("position", initial=0)
        v = phase.state("velocity", initial=0)
        u = phase.control("thrust")

        phase.dynamics({x: v, v: u})

        # Event constraints
        phase.event_constraints(
            x.final >= 5,  # Minimum final position
            v.final <= 2,  # Maximum final velocity
        )
        cost = phase.add_integral(u**2)
        problem.minimize(cost)  # Minimize control effort

        phase.mesh([4], [-1, 1])

        solution = mtor.solve_fixed_mesh(problem, nlp_options={"ipopt.print_level": 0})

        if solution.status["success"]:
            # Check event constraint satisfaction
            x_final = solution["position"][-1]
            v_final = solution["velocity"][-1]

            # Should satisfy event constraints (within tolerance)
            assert x_final >= 4.9  # Small tolerance
            assert v_final <= 2.1  # Small tolerance


class TestObjectiveFunctionTypes:
    def test_minimum_time_with_public_api(self):
        problem = mtor.Problem("Minimum Time Test")
        phase = problem.set_phase(1)

        t = phase.time(initial=0)  # Free final time
        x = phase.state("x", initial=0, final=1)
        u = phase.control("u", boundary=(-2, 2))

        phase.dynamics({x: u})
        problem.minimize(t.final)

        phase.mesh([4], [-1, 1])

        solution = mtor.solve_fixed_mesh(problem, nlp_options={"ipopt.print_level": 0})

        if solution.status["success"]:
            # Minimum time should be positive and reasonable
            objective = solution.status["objective"]
            assert objective > 0
            assert objective < 10  # Should be reasonable for this simple problem

    def test_integral_cost_with_public_api(self):
        problem = mtor.Problem("Integral Cost Test")
        phase = problem.set_phase(1)

        _t = phase.time(initial=0, final=1)
        x = phase.state("x", initial=0)
        u = phase.control("u")

        phase.dynamics({x: u})

        # Quadratic cost
        cost = phase.add_integral(x**2 + u**2)
        problem.minimize(cost)

        phase.mesh([4], [-1, 1])

        solution = mtor.solve_fixed_mesh(problem, nlp_options={"ipopt.print_level": 0})

        if solution.status["success"]:
            # Quadratic cost should be non-negative
            objective = solution.status["objective"]
            assert objective >= 0

    def test_parameter_dependent_objective_with_public_api(self):
        problem = mtor.Problem("Parameter Objective Test")

        weight = problem.parameter("weight", boundary=(0.1, 10))

        phase = problem.set_phase(1)
        x = phase.state("x", initial=0, final=1)
        u = phase.control("u")

        phase.dynamics({x: u})

        # Parameter-dependent objective
        cost = phase.add_integral(weight * u**2)
        problem.minimize(cost)

        phase.mesh([3], [-1, 1])

        solution = mtor.solve_fixed_mesh(problem, nlp_options={"ipopt.print_level": 0})

        if solution.status["success"]:
            assert solution.status["objective"] is not None

            params = solution.parameters
            if params is not None:
                assert len(params["values"]) == 1



from __future__ import annotations

from typing import Any, cast

import casadi as ca

from ..input_validation import _validate_constraint_input_format, _validate_string_not_empty
from .state import (
    ConstraintInput,
    MultiPhaseVariableState,
    PhaseDefinition,
    StaticParameterState,
    _BoundaryConstraint,
)


def _extract_casadi_symbol(expr: Any) -> ca.MX:
    if hasattr(expr, "_symbolic_var"):
        return expr._symbolic_var
    return expr


def _create_phase_symbol(base_name: str, phase_id: int, suffix: str = "") -> ca.MX:
    name = f"{base_name}_p{phase_id}{suffix}" if suffix else f"{base_name}_p{phase_id}"
    return ca.MX.sym(name, 1)  # type: ignore[arg-type]


def _validate_constraint_and_name(name: str, constraint: ConstraintInput, context: str) -> None:
    _validate_string_not_empty(name, f"{context} name")
    _validate_constraint_input_format(constraint, f"{context} '{name}' constraint")


def _convert_expression_to_casadi(
    expr: Any, expression_type: str, allow_callable_error: bool = True
) -> ca.MX:
    try:
        if isinstance(expr, ca.MX):
            return expr
        elif hasattr(expr, "_symbolic_var"):
            return _extract_casadi_symbol(expr)
        else:
            return ca.MX(expr)
    except Exception as e:
        if callable(expr) and allow_callable_error:
            raise ValueError(
                f"{expression_type} appears to be a function {expr}. Did you forget to call it?"
            ) from e
        else:
            raise ValueError(
                f"Cannot convert {expression_type} of type {type(expr)} to CasADi MX: {expr}"
            ) from e


def _validate_dynamics_key_exists(
    state_sym: ca.MX | StateVariableImpl, ordered_state_symbols: list[ca.MX], phase_id: int
) -> None:
    underlying_sym = _extract_casadi_symbol(state_sym)

    for sym in ordered_state_symbols:
        if underlying_sym is sym:
            return

    raise ValueError(f"Dynamics provided for undefined state variable in phase {phase_id}")


def _convert_dynamics_dict_to_casadi(
    dynamics_dict: dict[ca.MX | Any, ca.MX | float | int | Any], phase_id: int
) -> dict[ca.MX, ca.MX]:
    converted_dict = {}

    for key, value in dynamics_dict.items():
        storage_key = _extract_casadi_symbol(key)
        storage_value = _convert_expression_to_casadi(value, "Dynamics expression")
        converted_dict[storage_key] = storage_value

    return converted_dict


class _SymbolicVariableBase:
    def __init__(self, symbolic_var: ca.MX) -> None:
        self._symbolic_var = symbolic_var

    def __call__(self, other: Any = None) -> ca.MX:
        if other is None:
            return self._symbolic_var
        raise NotImplementedError("Variable indexing not yet implemented")

    def __casadi_MX__(self) -> ca.MX:  # noqa: N802
        return self._symbolic_var

    def __array_function__(self, func, args, kwargs):
        converted_args = [self._symbolic_var if arg is self else arg for arg in args]
        return func(*converted_args, **kwargs)

    def __getattr__(self, name: str) -> Any:
        if name.startswith("_"):
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")
        return getattr(self._symbolic_var, name)

    def __hash__(self) -> int:
        return hash(self._symbolic_var)

    def __eq__(self, other: Any) -> bool:
        if hasattr(other, "_symbolic_var"):
            return self._symbolic_var is other._symbolic_var
        return self._symbolic_var is other

    def __add__(self, other: Any) -> ca.MX:
        return self._symbolic_var + other

    def __radd__(self, other: Any) -> ca.MX:
        return other + self._symbolic_var

    def __sub__(self, other: Any) -> ca.MX:
        return self._symbolic_var - other

    def __rsub__(self, other: Any) -> ca.MX:
        return other - self._symbolic_var

    def __mul__(self, other: Any) -> ca.MX:
        return self._symbolic_var * other

    def __rmul__(self, other: Any) -> ca.MX:
        return other * self._symbolic_var

    def __truediv__(self, other: Any) -> ca.MX:
        return self._symbolic_var / other

    def __rtruediv__(self, other: Any) -> ca.MX:
        return other / self._symbolic_var

    def __pow__(self, other: Any) -> ca.MX:
        return self._symbolic_var**other

    def __neg__(self) -> ca.MX:
        return cast(ca.MX, -self._symbolic_var)

    def __lt__(self, other: Any) -> ca.MX:
        return self._symbolic_var < other

    def __le__(self, other: Any) -> ca.MX:
        return self._symbolic_var <= other

    def __gt__(self, other: Any) -> ca.MX:
        return self._symbolic_var > other

    def __ge__(self, other: Any) -> ca.MX:
        return self._symbolic_var >= other

    def __ne__(self, other: Any) -> ca.MX:
        return self._symbolic_var != other


class TimeVariableImpl(_SymbolicVariableBase):
    def __init__(self, sym_var: ca.MX, sym_initial: ca.MX, sym_final: ca.MX) -> None:
        super().__init__(sym_var)
        self._sym_initial = sym_initial
        self._sym_final = sym_final

    @property
    def initial(self) -> ca.MX:
        return self._sym_initial

    @property
    def final(self) -> ca.MX:
        return self._sym_final


class StateVariableImpl(_SymbolicVariableBase):
    def __init__(self, sym_var: ca.MX, sym_initial: ca.MX, sym_final: ca.MX) -> None:
        super().__init__(sym_var)
        self._sym_initial = sym_initial
        self._sym_final = sym_final

    @property
    def initial(self) -> ca.MX:
        return self._sym_initial

    @property
    def final(self) -> ca.MX:
        return self._sym_final


def _create_time_symbols(phase_id: int) -> tuple[ca.MX, ca.MX, ca.MX]:
    sym_time = _create_phase_symbol("t", phase_id)
    sym_t0 = _create_phase_symbol("t0", phase_id)
    sym_tf = _create_phase_symbol("tf", phase_id)
    return sym_time, sym_t0, sym_tf


def _create_state_symbols(name: str, phase_id: int) -> tuple[ca.MX, ca.MX, ca.MX]:
    sym_var = _create_phase_symbol(name, phase_id)
    sym_initial = _create_phase_symbol(f"{name}_initial", phase_id)
    sym_final = _create_phase_symbol(f"{name}_final", phase_id)
    return sym_var, sym_initial, sym_final


def create_phase_time_variable(
    phase_def: PhaseDefinition, initial: ConstraintInput = 0.0, final: ConstraintInput = None
) -> TimeVariableImpl:
    _validate_constraint_input_format(initial, f"phase {phase_def.phase_id} initial time")
    _validate_constraint_input_format(final, f"phase {phase_def.phase_id} final time")

    sym_time, sym_t0, sym_tf = _create_time_symbols(phase_def.phase_id)

    if initial is None:
        initial = 0.0

    t0_constraint = _BoundaryConstraint(initial)
    tf_constraint = _BoundaryConstraint(final)

    phase_def.t0_constraint = t0_constraint
    phase_def.tf_constraint = tf_constraint
    phase_def.sym_time = sym_time
    phase_def.sym_time_initial = sym_t0
    phase_def.sym_time_final = sym_tf

    return TimeVariableImpl(sym_time, sym_t0, sym_tf)


def _create_phase_state_variable(
    phase_def: PhaseDefinition,
    name: str,
    initial: ConstraintInput = None,
    final: ConstraintInput = None,
    boundary: ConstraintInput = None,
) -> StateVariableImpl:
    _validate_constraint_and_name(
        name, initial, f"phase {phase_def.phase_id} state '{name}' initial"
    )
    _validate_constraint_and_name(name, final, f"phase {phase_def.phase_id} state '{name}' final")
    _validate_constraint_and_name(
        name, boundary, f"phase {phase_def.phase_id} state '{name}' boundary"
    )

    sym_var, sym_initial, sym_final = _create_state_symbols(name, phase_def.phase_id)

    initial_constraint = _BoundaryConstraint(initial) if initial is not None else None
    final_constraint = _BoundaryConstraint(final) if final is not None else None
    boundary_constraint = _BoundaryConstraint(boundary) if boundary is not None else None

    phase_def.add_state(
        name=name,
        symbol=sym_var,
        initial_symbol=sym_initial,
        final_symbol=sym_final,
        initial_constraint=initial_constraint,
        final_constraint=final_constraint,
        boundary_constraint=boundary_constraint,
    )

    return StateVariableImpl(sym_var, sym_initial, sym_final)


def create_phase_control_variable(
    phase_def: PhaseDefinition, name: str, boundary: ConstraintInput = None
) -> ca.MX:
    _validate_constraint_and_name(name, boundary, f"phase {phase_def.phase_id} control '{name}'")

    sym_var = _create_phase_symbol(name, phase_def.phase_id)
    boundary_constraint = _BoundaryConstraint(boundary) if boundary is not None else None

    phase_def.add_control(name=name, symbol=sym_var, boundary_constraint=boundary_constraint)

    return sym_var


def _create_static_parameter(
    static_params: StaticParameterState, name: str, boundary: ConstraintInput = None
) -> ca.MX:
    _validate_constraint_and_name(name, boundary, f"parameter '{name}'")

    sym_var = ca.MX.sym(f"param_{name}", 1)  # type: ignore[arg-type]
    boundary_constraint = _BoundaryConstraint(boundary) if boundary is not None else None

    static_params.add_parameter(name=name, symbol=sym_var, boundary_constraint=boundary_constraint)

    return sym_var


def _set_phase_dynamics(
    phase_def: PhaseDefinition,
    dynamics_dict: dict[ca.MX | StateVariableImpl, ca.MX | float | int | StateVariableImpl],
) -> None:
    ordered_state_symbols = phase_def._get_ordered_state_symbols()

    for state_sym in dynamics_dict.keys():
        _validate_dynamics_key_exists(state_sym, ordered_state_symbols, phase_def.phase_id)

    converted_dict = _convert_dynamics_dict_to_casadi(dynamics_dict, phase_def.phase_id)
    phase_def.dynamics_expressions = converted_dict

    phase_def._dynamics_function = None
    phase_def._numerical_dynamics_function = None
    phase_def._functions_built = False


def _create_integral_symbol(phase_def: PhaseDefinition) -> ca.MX:
    integral_name = f"integral_{len(phase_def.integral_expressions)}_p{phase_def.phase_id}"
    return ca.MX.sym(integral_name, 1)  # type: ignore[arg-type]


def _set_phase_integral(phase_def: PhaseDefinition, integrand_expr: ca.MX | float | int) -> ca.MX:
    integral_sym = _create_integral_symbol(phase_def)
    pure_expr = _convert_expression_to_casadi(integrand_expr, "Integrand")

    phase_def.integral_expressions.append(pure_expr)
    phase_def.integral_symbols.append(integral_sym)
    phase_def.num_integrals = len(phase_def.integral_expressions)

    phase_def._integrand_function = None
    phase_def._functions_built = False

    return integral_sym


def _set_multiphase_objective(
    multiphase_state: MultiPhaseVariableState, objective_expr: ca.MX | float | int
) -> None:
    pure_expr = _convert_expression_to_casadi(objective_expr, "Objective")
    multiphase_state.objective_expression = pure_expr

    multiphase_state._objective_function = None
    multiphase_state._functions_built = False



import casadi as ca

from ..problem.state import PhaseDefinition


def _map_symbol_to_vector(symbol: ca.MX, vector: ca.MX, index: int, vector_length: int) -> ca.MX:
    return vector if vector_length == 1 else vector[index]


def _create_phase_symbol(base_name: str, phase_id: int, suffix: str = "") -> ca.MX:
    name = f"{base_name}_p{phase_id}{suffix}" if suffix else f"{base_name}_p{phase_id}"
    return ca.MX.sym(name, 1)  # type: ignore[arg-type]


def _create_state_symbols_for_phase(phase_id: int, num_states: int) -> tuple[ca.MX, ca.MX]:
    x0_vec = ca.vertcat(*[ca.MX.sym(f"x0_{i}_p{phase_id}", 1) for i in range(num_states)])  # type: ignore[arg-type]
    xf_vec = ca.vertcat(*[ca.MX.sym(f"xf_{i}_p{phase_id}", 1) for i in range(num_states)])  # type: ignore[arg-type]
    return ca.MX(x0_vec), ca.MX(xf_vec)


def _create_integral_symbol_for_phase(phase_id: int, num_integrals: int) -> ca.MX:
    if num_integrals > 0:
        return ca.vertcat(*[ca.MX.sym(f"q_{i}_p{phase_id}", 1) for i in range(num_integrals)])  # type: ignore[arg-type]
    return ca.MX.sym(f"q_p{phase_id}", 1)  # type: ignore[arg-type]


def _create_static_params_symbol(num_params: int) -> ca.MX:
    if num_params > 0:
        return ca.vertcat(*[ca.MX.sym(f"s_{i}", 1) for i in range(num_params)])  # type: ignore[arg-type]
    return ca.MX.sym("s", 1)  # type: ignore[arg-type]


def _build_unified_casadi_function_inputs(
    phase_def: PhaseDefinition,
    static_parameter_symbols: list[ca.MX] | None = None,
) -> tuple[ca.MX, ca.MX, ca.MX, ca.MX, list[ca.MX]]:
    state_syms = phase_def._get_ordered_state_symbols()
    control_syms = phase_def._get_ordered_control_symbols()

    states_vec = ca.vertcat(*state_syms) if state_syms else ca.MX()
    controls_vec = ca.vertcat(*control_syms) if control_syms else ca.MX()
    time = phase_def.sym_time if phase_def.sym_time is not None else ca.MX.sym("t", 1)  # type: ignore[arg-type]

    num_static_params = len(static_parameter_symbols) if static_parameter_symbols else 0
    static_params_vec = (
        ca.MX.sym("static_params", num_static_params, 1)  # type: ignore[arg-type]
        if num_static_params > 0
        else ca.MX.sym("static_params", 1, 1)  # type: ignore[arg-type]
    )

    function_inputs = [states_vec, controls_vec, time, static_params_vec]

    return (
        ca.MX(states_vec),
        ca.MX(controls_vec),
        ca.MX(time),
        ca.MX(static_params_vec),
        function_inputs,
    )


def _build_static_parameter_substitution_map(
    static_parameter_symbols: list[ca.MX] | None,
    static_params_vec: ca.MX,
) -> dict[ca.MX, ca.MX]:
    subs_map = {}
    num_static_params = len(static_parameter_symbols) if static_parameter_symbols else 0

    if static_parameter_symbols and num_static_params > 0:
        for i, param_sym in enumerate(static_parameter_symbols):
            subs_map[param_sym] = _map_symbol_to_vector(
                param_sym, static_params_vec, i, num_static_params
            )

    return subs_map


def _build_phase_inputs_for_phase(multiphase_state, phase_id: int) -> list[ca.MX]:
    phase_def = multiphase_state.phases[phase_id]

    t0 = (
        phase_def.sym_time_initial
        if phase_def.sym_time_initial is not None
        else _create_phase_symbol("t0", phase_id)
    )
    tf = (
        phase_def.sym_time_final
        if phase_def.sym_time_final is not None
        else _create_phase_symbol("tf", phase_id)
    )

    state_syms = phase_def._get_ordered_state_symbols()
    x0_vec, xf_vec = _create_state_symbols_for_phase(phase_id, len(state_syms))

    q_vec = _create_integral_symbol_for_phase(phase_id, phase_def.num_integrals)

    return [t0, tf, x0_vec, xf_vec, q_vec]


def _build_unified_multiphase_symbol_inputs(multiphase_state) -> tuple[list[ca.MX], ca.MX]:
    phase_inputs = []

    for phase_id in sorted(multiphase_state.phases.keys()):
        phase_inputs.extend(_build_phase_inputs_for_phase(multiphase_state, phase_id))

    static_param_syms = multiphase_state.static_parameters.get_ordered_parameter_symbols()
    s_vec = _create_static_params_symbol(len(static_param_syms))

    phase_inputs.append(s_vec)

    return phase_inputs, ca.MX(s_vec)


def _map_state_symbols_for_phase(
    phase_def: PhaseDefinition,
    x0_vec: ca.MX,
    xf_vec: ca.MX,
    phase_symbols_map: dict[ca.MX, ca.MX],
) -> None:
    state_syms = phase_def._get_ordered_state_symbols()
    state_initial_syms = phase_def._get_ordered_state_initial_symbols()
    state_final_syms = phase_def._get_ordered_state_final_symbols()
    num_states = len(state_syms)

    for i, (state_sym, initial_sym, final_sym) in enumerate(
        zip(state_syms, state_initial_syms, state_final_syms, strict=True)
    ):
        state_value = _map_symbol_to_vector(state_sym, xf_vec, i, num_states)
        initial_value = _map_symbol_to_vector(initial_sym, x0_vec, i, num_states)
        final_value = _map_symbol_to_vector(final_sym, xf_vec, i, num_states)

        phase_symbols_map[state_sym] = state_value
        phase_symbols_map[initial_sym] = initial_value
        phase_symbols_map[final_sym] = final_value


def _map_integral_symbols_for_phase(
    phase_def: PhaseDefinition, q_vec: ca.MX, phase_symbols_map: dict[ca.MX, ca.MX]
) -> None:
    for i, integral_sym in enumerate(phase_def.integral_symbols):
        phase_symbols_map[integral_sym] = _map_symbol_to_vector(
            integral_sym, q_vec, i, phase_def.num_integrals
        )


def _map_time_symbols_for_phase(
    t0: ca.MX, tf: ca.MX, phase_symbols_map: dict[ca.MX, ca.MX]
) -> None:
    phase_symbols_map[t0] = t0
    phase_symbols_map[tf] = tf


def _map_static_parameter_symbols(
    static_param_syms: list[ca.MX], s_vec: ca.MX, phase_symbols_map: dict[ca.MX, ca.MX]
) -> None:
    num_params = len(static_param_syms)
    for i, param_sym in enumerate(static_param_syms):
        phase_symbols_map[param_sym] = _map_symbol_to_vector(param_sym, s_vec, i, num_params)


def _build_unified_symbol_substitution_map(
    multiphase_state, phase_inputs: list[ca.MX], s_vec: ca.MX
) -> dict[ca.MX, ca.MX]:
    phase_symbols_map = {}
    input_idx = 0

    for phase_id in sorted(multiphase_state.phases.keys()):
        phase_def = multiphase_state.phases[phase_id]

        t0 = phase_inputs[input_idx]
        tf = phase_inputs[input_idx + 1]
        x0_vec = phase_inputs[input_idx + 2]
        xf_vec = phase_inputs[input_idx + 3]
        q_vec = phase_inputs[input_idx + 4]
        input_idx += 5

        _map_time_symbols_for_phase(t0, tf, phase_symbols_map)
        _map_state_symbols_for_phase(phase_def, x0_vec, xf_vec, phase_symbols_map)
        _map_integral_symbols_for_phase(phase_def, q_vec, phase_symbols_map)

    static_param_syms = multiphase_state.static_parameters.get_ordered_parameter_symbols()
    _map_static_parameter_symbols(static_param_syms, s_vec, phase_symbols_map)

    return phase_symbols_map



import logging
from collections.abc import Sequence
from typing import Any

import casadi as ca

from maptor.exceptions import ConfigurationError

from ..input_validation import (
    _validate_complete_dynamics,
    _validate_constraint_input_format,
    _validate_string_not_empty,
)
from ..mtor_types import NumericArrayLike, PhaseID
from . import constraints_problem, initial_guess_problem, mesh, solver_interface, variables_problem
from .constraints_problem import (
    _get_cross_phase_event_constraints_function,
    _get_phase_path_constraints_function,
)
from .state import ConstraintInput, MultiPhaseVariableState
from .variables_problem import StateVariableImpl, TimeVariableImpl


logger = logging.getLogger(__name__)


def _validate_phase_exists(phases: dict[PhaseID, Any], phase_id: PhaseID) -> None:
    if phase_id not in phases:
        raise ValueError(f"Phase {phase_id} does not exist")


def _validate_constraint_inputs(name: str, boundary: ConstraintInput, context: str) -> None:
    _validate_string_not_empty(name, f"{context} name")
    _validate_constraint_input_format(boundary, f"{context} '{name}' boundary")


def _log_constraint_addition(
    constraint_count: int, phase_id: PhaseID, constraint_type: str
) -> None:
    logger.debug(
        "Added %d %s constraint(s) to phase %d", constraint_count, constraint_type, phase_id
    )


def _validate_constraint_expressions_not_empty(
    constraint_expressions: tuple, phase_id: PhaseID, constraint_type: str
) -> None:
    if not constraint_expressions:
        raise ValueError(
            f"Phase {phase_id} {constraint_type}_constraints() requires at least one constraint expression"
        )


def _process_symbolic_time_constraints(
    phase_def: Any, phase_id: PhaseID, cross_phase_constraints: list[ca.MX]
) -> None:
    if phase_def.t0_constraint.is_symbolic():
        if (
            phase_def.sym_time_initial is None
            or phase_def.t0_constraint.symbolic_expression is None
        ):
            raise ValueError(f"Phase {phase_id} has an undefined symbolic time initial expression.")
        constraint_expr = phase_def.sym_time_initial - phase_def.t0_constraint.symbolic_expression
        cross_phase_constraints.append(constraint_expr)
        logger.debug(f"Added automatic time initial constraint for phase {phase_id}")

    if phase_def.tf_constraint.is_symbolic():
        if phase_def.sym_time_final is None or phase_def.tf_constraint.symbolic_expression is None:
            raise ValueError(f"Phase {phase_id} has an undefined symbolic time final expression.")
        constraint_expr = phase_def.sym_time_final - phase_def.tf_constraint.symbolic_expression
        cross_phase_constraints.append(constraint_expr)
        logger.debug(f"Added automatic time final constraint for phase {phase_id}")


def _process_single_symbolic_boundary_constraint(
    var_name: str,
    constraint_type: str,
    symbolic_expr: ca.MX,
    phase_def: Any,
    phase_id: PhaseID,
    cross_phase_constraints: list[ca.MX],
) -> None:
    state_index = phase_def.state_name_to_index[var_name]

    if constraint_type == "initial":
        state_initial_sym = phase_def._get_ordered_state_initial_symbols()[state_index]
        constraint_expr = state_initial_sym - symbolic_expr
        cross_phase_constraints.append(constraint_expr)
        logger.debug(f"Added automatic initial constraint for phase {phase_id} state '{var_name}'")

    elif constraint_type == "final":
        state_final_sym = phase_def._get_ordered_state_final_symbols()[state_index]
        constraint_expr = state_final_sym - symbolic_expr
        cross_phase_constraints.append(constraint_expr)
        logger.debug(f"Added automatic final constraint for phase {phase_id} state '{var_name}'")

    elif constraint_type == "boundary":
        state_initial_sym = phase_def._get_ordered_state_initial_symbols()[state_index]
        state_final_sym = phase_def._get_ordered_state_final_symbols()[state_index]

        initial_constraint = state_initial_sym - symbolic_expr
        final_constraint = state_final_sym - symbolic_expr

        cross_phase_constraints.extend([initial_constraint, final_constraint])
        logger.debug(
            f"Added automatic boundary constraints for phase {phase_id} state '{var_name}'"
        )


def _process_symbolic_state_constraints(
    phase_def: Any, phase_id: PhaseID, cross_phase_constraints: list[ca.MX]
) -> None:
    for var_name, constraint_type, symbolic_expr in phase_def.symbolic_boundary_constraints:
        _process_single_symbolic_boundary_constraint(
            var_name, constraint_type, symbolic_expr, phase_def, phase_id, cross_phase_constraints
        )


def _validate_phase_requirements(phases: dict[PhaseID, Any]) -> None:
    if not phases:
        raise ConfigurationError("Problem must have at least one phase defined")

    for phase_id, phase_def in phases.items():
        if not phase_def.dynamics_expressions:
            raise ConfigurationError(f"Phase {phase_id} must have dynamics defined")
        _validate_complete_dynamics(phase_def, phase_id)
        if not phase_def.mesh_configured:
            raise ConfigurationError(
                f"Phase {phase_id} must have mesh configured before validation. "
                f"Call phase.mesh(polynomial_degrees, mesh_points) first."
            )


def _process_symbolic_constraints_for_all_phases(
    phases: dict[PhaseID, Any], cross_phase_constraints: list[ca.MX]
) -> None:
    logger.debug("Processing symbolic boundary constraints for automatic cross-phase linking")

    for phase_id, phase_def in phases.items():
        _process_symbolic_time_constraints(phase_def, phase_id, cross_phase_constraints)
        _process_symbolic_state_constraints(phase_def, phase_id, cross_phase_constraints)


def _build_all_phase_functions(multiphase_state: MultiPhaseVariableState) -> None:
    static_parameter_symbols = multiphase_state.static_parameters.get_ordered_parameter_symbols()

    logger.debug("Building functions for all phases")
    for phase_id, phase_def in multiphase_state.phases.items():
        if phase_def._functions_built:
            continue

        logger.debug(f"Building functions for phase {phase_id}")

        phase_def._dynamics_function, phase_def._numerical_dynamics_function = (
            solver_interface._build_unified_phase_dynamics_functions(
                phase_def, static_parameter_symbols
            )
        )

        phase_def._integrand_function = solver_interface._build_phase_integrand_function(
            phase_def, static_parameter_symbols
        )

        phase_def._path_constraints_function = _get_phase_path_constraints_function(phase_def)

        phase_def._functions_built = True

    if not multiphase_state._functions_built:
        logger.debug("Building multiphase objective function")
        multiphase_state._objective_function = (
            solver_interface._build_multiphase_objective_function(multiphase_state)
        )

        multiphase_state._cross_phase_constraints_function = (
            _get_cross_phase_event_constraints_function(multiphase_state)
        )

        multiphase_state._functions_built = True


class Phase:
    """
    Single phase definition for multiphase optimal control problems.

    A Phase represents one segment of a multiphase trajectory with its own time
    domain, state variables, control inputs, dynamics, and constraints. Phases
    can be linked together through symbolic constraints to create complex
    multiphase missions.

    The Phase class provides a fluent interface for defining:

    - Time variables with boundary conditions
    - State variables with initial, final, and path constraints
    - Control variables with bounds
    - System dynamics as differential equations
    - Integral cost terms and constraints
    - Path constraints applied throughout the phase
    - Event constraints at phase boundaries
    - Mesh discretization for numerical solution

    Note:
        Phase objects are created through Problem.set_phase() and should not
        be instantiated directly.

    Examples:
        Basic single-phase problem setup:

        >>> problem = mtor.Problem("Rocket Ascent")
        >>> phase = problem.set_phase(1)
        >>>
        >>> # Define time and variables
        >>> t = phase.time(initial=0, final=10)
        >>> h = phase.state("altitude", initial=0, final=1000)
        >>> v = phase.state("velocity", initial=0)
        >>> T = phase.control("thrust", boundary=(0, 2000))
        >>>
        >>> # Set dynamics
        >>> phase.dynamics({h: v, v: T/1000 - 9.81})
        >>>
        >>> # Add constraints and mesh
        >>> phase.path_constraints(h >= 0, T <= 1500)
        >>> phase.mesh([5, 5], [-1, 0, 1])

        Multiphase trajectory with automatic linking:

        >>> # Phase 1: Boost
        >>> p1 = problem.set_phase(1)
        >>> t1 = p1.time(initial=0, final=120)
        >>> h1 = p1.state("altitude", initial=0)
        >>> v1 = p1.state("velocity", initial=0)
        >>> # ... dynamics and constraints
        >>>
        >>> # Phase 2: Coast (automatically linked)
        >>> p2 = problem.set_phase(2)
        >>> t2 = p2.time(initial=t1.final, final=300)  # Continuous time
        >>> h2 = p2.state("altitude", initial=h1.final)  # Continuous altitude
        >>> v2 = p2.state("velocity", initial=v1.final)  # Continuous velocity
        >>> # ... dynamics for coast phase
    """

    def __init__(self, problem: "Problem", phase_id: PhaseID) -> None:
        self.problem = problem
        self.phase_id = phase_id
        self._phase_def = self.problem._multiphase_state.set_phase(self.phase_id)

    def time(
        self, initial: ConstraintInput = 0.0, final: ConstraintInput = None
    ) -> TimeVariableImpl:
        """
        Define the time variable for this phase with boundary conditions.

        Creates the time coordinate for this phase with comprehensive constraint
        specification. Supports fixed times, bounded ranges, and symbolic expressions
        for multiphase trajectory continuity.

        Args:
            initial: Initial time constraint with full constraint syntax:

                - float: Fixed initial time (e.g., 0.0)
                - (lower, upper): Bounded initial time range (e.g., (0, 10))
                - (None, upper): Upper bounded only (e.g., (None, 5))
                - (lower, None): Lower bounded only (e.g., (2, None))
                - ca.MX: Symbolic expression for phase linking
                - None: Unconstrained initial time (optimization variable)

            final: Final time constraint with full constraint syntax:

                - float: Fixed final time (e.g., 100.0)
                - (lower, upper): Bounded final time range (e.g., (90, 110))
                - (None, upper): Upper bounded only (e.g., (None, 200))
                - (lower, None): Lower bounded only (e.g., (50, None))
                - ca.MX: Symbolic expression for phase linking
                - None: Free final time (optimization variable)

        Returns:
            TimeVariableImpl: Time variable with .initial and .final properties

        Examples:
            Fixed values:

            >>> t = phase.time(initial=0.0, final=10.0)

            Ranges:

            >>> t = phase.time(initial=(0, 5), final=(8, 12))

            Single-sided bounds:

            >>> t = phase.time(initial=(None, 5), final=(10, None))

            Free variables:

            >>> t = phase.time()  # Both free

            Symbolic linking:

            >>> t2 = phase2.time(initial=t1.final)
        """
        return variables_problem.create_phase_time_variable(self._phase_def, initial, final)

    def state(
        self,
        name: str,
        initial: ConstraintInput = None,
        final: ConstraintInput = None,
        boundary: ConstraintInput = None,
    ) -> StateVariableImpl:
        """
        Define a state variable with comprehensive constraint specification.

        Creates a state variable with exhaustive constraint capabilities including
        boundary conditions, path bounds, and symbolic multiphase linking. All
        constraint types support the full constraint syntax.

        Args:
            name: Unique state variable name within this phase
            initial: Initial state constraint with full syntax:

                - float: Fixed initial value (e.g., 0.0)
                - (lower, upper): Bounded initial range (e.g., (0, 10))
                - (None, upper): Upper bounded only (e.g., (None, 100))
                - (lower, None): Lower bounded only (e.g., (50, None))
                - ca.MX: Symbolic expression for multiphase continuity
                - None: Unconstrained initial state

            final: Final state constraint with full syntax:

                - float: Fixed final value (e.g., 1000.0)
                - (lower, upper): Bounded final range (e.g., (990, 1010))
                - (None, upper): Upper bounded only (e.g., (None, 1200))
                - (lower, None): Lower bounded only (e.g., (800, None))
                - ca.MX: Symbolic expression for multiphase continuity
                - None: Unconstrained final state

            boundary: Path constraint applied throughout trajectory:

                - float: State equals this value at all times
                - (lower, upper): State bounds throughout (e.g., (0, 1000))
                - (None, upper): Upper path bound only (e.g., (None, 500))
                - (lower, None): Lower path bound only (e.g., (0, None))
                - None: No path bounds

            Returns:
                StateVariableImpl: State variable with .initial and .final properties

            Examples:
                Fixed values:

                >>> altitude = phase.state("altitude", initial=0.0, final=1000.0)

            Ranges:

            >>> position = phase.state("position", initial=(0, 5), final=(95, 105))

            Single-sided bounds:

            >>> velocity = phase.state("velocity", initial=0, final=(None, 200))

            Path bounds:

            >>> mass = phase.state("mass", boundary=(100, 1000))

            All constraint types:

            >>> state = phase.state("x", initial=(0, 10), final=(90, 110), boundary=(0, None))

            Symbolic linking:

            >>> h2 = phase2.state("altitude", initial=h1.final)

            Unconstrained:

            >>> free_state = phase.state("free_variable")
        """
        return variables_problem._create_phase_state_variable(
            self._phase_def, name, initial, final, boundary
        )

    def control(self, name: str, boundary: ConstraintInput = None) -> ca.MX:
        """
        Define a control variable with comprehensive bound specification.

        Creates a control input with exhaustive constraint capabilities. Control
        variables represent actuator commands that can be optimized subject to
        physical or design limitations.

        Args:
            name: Unique control variable name within this phase
            boundary: Control bounds with full constraint syntax:

                - float: Fixed control value (e.g., 100.0)
                - (lower, upper): Symmetric/asymmetric bounds (e.g., (-50, 100))
                - (None, upper): Upper bound only (e.g., (None, 1000))
                - (lower, None): Lower bound only (e.g., (0, None))
                - None: Unconstrained control

        Returns:
            ca.MX: Control variable for use in dynamics and cost functions

        Examples:
            Bounded control:

            >>> thrust = phase.control("thrust", boundary=(0, 2000))

            Single-sided bounds:

            >>> power = phase.control("power", boundary=(0, None))
            >>> brake = phase.control("brake", boundary=(None, 100))

            Fixed value:

            >>> constant = phase.control("thrust", boundary=1500)

            Unconstrained:

            >>> free_control = phase.control("force")
        """
        return variables_problem.create_phase_control_variable(self._phase_def, name, boundary)

    def dynamics(
        self,
        dynamics_dict: dict[ca.MX | StateVariableImpl, ca.MX | float | int | StateVariableImpl],
    ) -> None:
        """
        Define differential equations with comprehensive expression support.

        Specifies system ODEs describing state evolution. Dynamics expressions
        can involve states, controls, time, parameters, and arbitrary mathematical
        relationships. Must provide dynamics for all states in the phase.

        Args:
            dynamics_dict: Maps state variables to time derivatives (dx/dt).
                Keys: State variables from state() method
                Values: Expressions for derivatives using states, controls, parameters

        Examples:
            Basic dynamics:

            >>> pos = phase.state("position")
            >>> vel = phase.state("velocity")
            >>> thrust = phase.control("thrust")
            >>> phase.dynamics({pos: vel, vel: thrust - 0.1*vel})

            With mass:

            >>> mass = phase.state("mass", initial=1000)
            >>> phase.dynamics({
            ...     pos: vel,
            ...     vel: thrust/mass - 9.81,
            ...     mass: -thrust * 0.001
            ... })

            With parameters:

            >>> drag_coeff = problem.parameter("drag")
            >>> phase.dynamics({pos: vel, vel: thrust - drag_coeff*vel**2})

            Multiple states/controls:

            >>> phase.dynamics({
            ...     x: vx,
            ...     y: vy,
            ...     vx: fx,
            ...     vy: fy - 9.81
            ... })
        """
        self._phase_def._functions_built = False
        variables_problem._set_phase_dynamics(self._phase_def, dynamics_dict)
        logger.info(
            "Dynamics defined for phase %d with %d state variables",
            self.phase_id,
            len(dynamics_dict),
        )

    def add_integral(self, integrand_expr: ca.MX | float | int) -> ca.MX:
        """
        Add integral term with comprehensive integrand expression support.

        Creates integral variables for cost functions, constraint integrals, and
        accumulated quantities. Supports arbitrary expressions involving states,
        controls, time, and parameters.

        Args:
            integrand_expr: Expression to integrate over phase duration.
                Can involve any combination of states, controls, time, parameters.

        Returns:
            ca.MX: Integral variable for use in objectives and constraints

        Examples:
            Quadratic cost:

            >>> cost = phase.add_integral(x**2 + u**2)
            >>> problem.minimize(cost)

            Resource consumption:

            >>> fuel = phase.add_integral(thrust * 0.001)

            Distance traveled:

            >>> distance = phase.add_integral(velocity)

            Multiple integrals:

            >>> energy = phase.add_integral(thrust**2)
            >>> fuel = phase.add_integral(thrust * rate)
            >>> problem.minimize(energy + 10*fuel)
        """
        self._phase_def._functions_built = False
        return variables_problem._set_phase_integral(self._phase_def, integrand_expr)

    def path_constraints(self, *constraint_expressions: ca.MX | float | int) -> None:
        r"""
        Add path constraints enforced continuously throughout the trajectory.

        Path constraints are enforced at all collocation points, ensuring conditions
        hold throughout the phase. Use for safety limits, physical bounds, and
        trajectory shaping that cannot be expressed through state boundary parameters.

        Args:
            \*constraint_expressions: Constraint expressions enforced continuously

        Examples:
            State bounds:

            >>> phase.path_constraints(altitude >= 0, velocity <= 250)

            Control bounds:

            >>> phase.path_constraints(thrust >= 0, thrust <= 2000)

            Complex expressions:

            >>> phase.path_constraints((x-50)**2 + (y-50)**2 >= 100)

            Multiple constraints:

            >>> phase.path_constraints(
            ...     altitude >= 0,
            ...     velocity <= 200,
            ...     acceleration <= 20
            ... )
        """
        _validate_constraint_expressions_not_empty(constraint_expressions, self.phase_id, "path")

        self._phase_def._functions_built = False
        for expr in constraint_expressions:
            constraints_problem._add_path_constraint(self._phase_def, expr)

        _log_constraint_addition(len(constraint_expressions), self.phase_id, "path")

    def event_constraints(self, *constraint_expressions: ca.MX | float | int) -> None:
        r"""
        Add event constraints for expressions not representable as state/control bounds.

        **Primary Purpose:** Constraints on integral terms, static parameters,
        cross-phase expressions, and complex mathematical relationships that cannot
        be expressed through state initial/final/boundary or control boundary parameters.

        **When to use event_constraints vs state parameters:**
        - Use state parameters for simple state boundaries: `state("x", final=(90, 110))`
        - Use event_constraints for integrals, parameters, and complex expressions

        Args:
            \*constraint_expressions: Constraint expressions involving:

                - Integral terms from `phase.add_integral()`
                - Static parameters from `problem.parameter()`
                - Cross-phase linking expressions
                - Complex mathematical relationships
                - Multi-variable constraint expressions

        Examples:
            Integral constraints:

            >>> fuel_used = phase.add_integral(thrust * 0.001)
            >>> phase.event_constraints(fuel_used <= 100)

            Parameter constraints:

            >>> mass = problem.parameter("mass", boundary=(100, 1000))
            >>> phase.event_constraints(mass >= 200)

            Cross-phase linking:

            >>> phase2.event_constraints(h2.initial == h1.final)

            Complex expressions:

            >>> phase.event_constraints(x.final**2 + y.final**2 >= 100)
        """
        _validate_constraint_expressions_not_empty(constraint_expressions, self.phase_id, "event")

        for expr in constraint_expressions:
            constraints_problem._add_event_constraint(self.problem._multiphase_state, expr)

        _log_constraint_addition(len(constraint_expressions), self.phase_id, "event")

    def mesh(self, polynomial_degrees: list[int], mesh_points: NumericArrayLike) -> None:
        """
        Configure pseudospectral mesh with comprehensive discretization control.

        Defines mesh discretization for Radau pseudospectral method with precise
        control over polynomial degrees and interval distribution. Critical for
        balancing solution accuracy with computational efficiency.

        Args:
            polynomial_degrees: Polynomial degree for each mesh interval.
                Higher degrees: better accuracy, more computational cost.
                Typical range: 3-12, up to 15 for very smooth solutions.
            mesh_points: Normalized mesh points in [-1, 1] defining interval
                boundaries. Length must equal len(polynomial_degrees) + 1.
                Points automatically scaled to actual phase time domain.

        Examples:
            Uniform mesh:

            >>> phase.mesh([4, 4, 4], [-1, -1/3, 1/3, 1])

            Non-uniform intervals:

            >>> phase.mesh([6, 4], [-1, -0.5, 1])

            High accuracy:

            >>> phase.mesh([10, 10], [-1, 0, 1])

            Single interval:

            >>> phase.mesh([5], [-1, 1])
        """

        logger.info(
            "Setting mesh for phase %d: %d intervals", self.phase_id, len(polynomial_degrees)
        )
        mesh._configure_phase_mesh(self._phase_def, polynomial_degrees, mesh_points)

    def guess(
        self,
        states: Sequence[NumericArrayLike] | None = None,
        controls: Sequence[NumericArrayLike] | None = None,
        initial_time: float | None = None,
        terminal_time: float | None = None,
        integrals: float | NumericArrayLike | None = None,
    ) -> None:
        """
        Provide initial guess for this phase's optimization variables.

        Supplies starting values for NLP solver with phase-specific variable coverage.
        Arrays must match this phase's mesh configuration exactly.

        Args:
            states: State trajectory guess per mesh interval.
                Structure: [interval_arrays]
                Each interval_array: shape (num_states, num_collocation_points)

            controls: Control trajectory guess per mesh interval.
                Structure: [interval_arrays]
                Each interval_array: shape (num_controls, num_mesh_points)

            initial_time: Initial time guess for this phase.

            terminal_time: Final time guess for this phase.

            integrals: Integral value guess for this phase.
                - float: Single integral
                - array: Multiple integrals

        Examples:
            Basic single phase:

            >>> states_guess = [np.array([[0, 1, 2, 3, 4], [0, 0, 1, 2, 3]])]
            >>> controls_guess = [np.array([[1, 1, 1, 1]])]
            >>> phase.guess(states=states_guess, controls=controls_guess, terminal_time=10.0)

            With integrals:

            >>> phase.guess(states=states_guess, integrals=50.0)
        """
        components = []
        if states is not None:
            components.append("states")
        if controls is not None:
            components.append("controls")
        if initial_time is not None:
            components.append("initial_time")
        if terminal_time is not None:
            components.append("terminal_time")
        if integrals is not None:
            components.append("integrals")

        logger.info("Setting initial guess for phase %d: %s", self.phase_id, ", ".join(components))

        # Reuse existing validation logic
        initial_guess_problem._set_phase_initial_guess(
            self._phase_def,
            states=states,
            controls=controls,
            initial_time=initial_time,
            terminal_time=terminal_time,
            integrals=integrals,
        )


class Problem:
    """
    Multiphase optimal control problem definition and configuration interface.

    The Problem class is the main entry point for defining optimal control problems
    in MAPTOR. It supports both single-phase and multiphase trajectory optimization
    with automatic phase linking, static parameter optimization, and comprehensive
    constraint specification.

    Key capabilities:

    - **Multiphase trajectory definition** with automatic continuity
    - **Static parameter optimization** (design variables)
    - **Flexible objective functions** (minimize/maximize any expression)
    - **Initial guess specification** for improved convergence
    - **Solver configuration** and validation
    - **Cross-phase constraints** and event handling

    The Problem follows a builder pattern where you incrementally define phases,
    variables, dynamics, constraints, and objectives before solving.

    Examples:
        Single-phase minimum time problem:

        >>> import maptor as mtor
        >>>
        >>> problem = mtor.Problem("Minimum Time")
        >>> phase = problem.set_phase(1)
        >>>
        >>> # Variables and dynamics
        >>> t = phase.time(initial=0.0)
        >>> x = phase.state("position", initial=0, final=1)
        >>> v = phase.state("velocity", initial=0)
        >>> u = phase.control("force", boundary=(-1, 1))
        >>>
        >>> phase.dynamics({x: v, v: u})
        >>> problem.minimize(t.final)
        >>>
        >>> # Solve
        >>> phase.mesh([5, 5], [-1, 0, 1])
        >>> solution = mtor.solve_fixed_mesh(problem)

        Multiphase rocket trajectory:

        >>> problem = mtor.Problem("Rocket Launch")
        >>>
        >>> # Boost phase
        >>> boost = problem.set_phase(1)
        >>> t1 = boost.time(initial=0, final=120)
        >>> h1 = boost.state("altitude", initial=0)
        >>> v1 = boost.state("velocity", initial=0)
        >>> m1 = boost.state("mass", initial=1000)
        >>> T1 = boost.control("thrust", boundary=(0, 2000))
        >>>
        >>> boost.dynamics({
        ...     h1: v1,
        ...     v1: T1/m1 - 9.81,
        ...     m1: -T1 * 0.001
        ... })
        >>>
        >>> # Coast phase with automatic continuity
        >>> coast = problem.set_phase(2)
        >>> t2 = coast.time(initial=t1.final, final=300)
        >>> h2 = coast.state("altitude", initial=h1.final)
        >>> v2 = coast.state("velocity", initial=v1.final)
        >>> m2 = coast.state("mass", initial=m1.final)
        >>>
        >>> coast.dynamics({h2: v2, v2: -9.81, m2: 0})
        >>>
        >>> # Objective and solve
        >>> problem.minimize(-h2.final)  # Maximize final altitude
        >>> # ... mesh configuration and solve

        Problem with static parameters:

        >>> problem = mtor.Problem("Design Optimization")
        >>>
        >>> # Design parameters to optimize
        >>> engine_mass = problem.parameter("engine_mass", boundary=(50, 200))
        >>> thrust_level = problem.parameter("max_thrust", boundary=(1000, 5000))
        >>>
        >>> # Use parameters in dynamics
        >>> total_mass = vehicle_mass + engine_mass
        >>> phase.dynamics({v: thrust_level/total_mass - 9.81})
        >>>
        >>> # Multi-objective: maximize performance, minimize mass
        >>> performance = altitude.final
        >>> mass_penalty = engine_mass * 10
        >>> problem.minimize(mass_penalty - performance)
    """

    def __init__(self, name: str = "Multiphase Problem") -> None:
        """
        Initialize a new optimal control problem.

        Args:
            name: Descriptive name for the problem (used in logging and output)

        Examples:
            >>> problem = mtor.Problem("Spacecraft Trajectory")
            >>> problem = mtor.Problem("Robot Path Planning")
            >>> problem = mtor.Problem()  # Uses default name
        """
        _validate_string_not_empty(name, "Problem name")
        self.name = name
        logger.debug("Created multiphase problem: '%s'", name)

        self._multiphase_state = MultiPhaseVariableState()
        self.solver_options: dict[str, Any] = {}

    def set_phase(self, phase_id: PhaseID) -> Phase:
        """
        Create and configure a new phase in the multiphase problem.

        Each phase represents a distinct segment of the trajectory with its own
        time domain, dynamics, and constraints. Phases can be linked through
        symbolic boundary constraints for trajectory continuity.

        Args:
            phase_id: Unique integer identifier for this phase. Phases are solved
                in order of their IDs, so use sequential numbering (1, 2, 3...).

        Returns:
            Phase: Phase object for defining variables, dynamics, and constraints

        Raises:
            ValueError: If phase_id already exists

        Examples:
            Single phase problem:

            >>> problem = mtor.Problem("Single Phase")
            >>> phase = problem.set_phase(1)
            >>> # Define phase variables, dynamics, constraints...

            Sequential multiphase problem:

            >>> problem = mtor.Problem("Three Phase Mission")
            >>>
            >>> # Launch phase
            >>> launch = problem.set_phase(1)
            >>> # ... configure launch phase
            >>>
            >>> # Coast phase
            >>> coast = problem.set_phase(2)
            >>> # ... configure coast phase
            >>>
            >>> # Landing phase
            >>> landing = problem.set_phase(3)
            >>> # ... configure landing phase

            Phase naming convention:

            >>> # Use descriptive variable names for clarity
            >>> ascent_phase = problem.set_phase(1)
            >>> orbit_phase = problem.set_phase(2)
            >>> descent_phase = problem.set_phase(3)
        """
        if phase_id in self._multiphase_state.phases:
            raise ValueError(f"Phase {phase_id} already exists")

        logger.debug("Adding phase %d to problem '%s'", phase_id, self.name)
        return Phase(self, phase_id)

    def parameter(self, name: str, boundary: ConstraintInput = None) -> ca.MX:
        """
        Define a static parameter for design optimization with exhaustive constraint syntax.

        Creates optimization variables that remain constant throughout the mission
        but can be varied to optimize performance. Supports all constraint types
        for comprehensive design space specification.

        Args:
            name: Unique parameter name
            boundary: Parameter constraint with full syntax:

                - float: Fixed parameter value (e.g., 1000.0)
                - (lower, upper): Bounded parameter range (e.g., (100, 500))
                - (None, upper): Upper bounded only (e.g., (None, 1000))
                - (lower, None): Lower bounded only (e.g., (0, None))
                - None: Unconstrained parameter

        Returns:
            ca.MX: Parameter variable for use across all phases

        Examples:
            Bounded parameter:

            >>> mass = problem.parameter("mass", boundary=(100, 500))

            Single-sided bounds:

            >>> area = problem.parameter("area", boundary=(10, None))
            >>> drag = problem.parameter("drag", boundary=(None, 0.5))

            Fixed parameter:

            >>> gravity = problem.parameter("gravity", boundary=9.81)

            Unconstrained:

            >>> free_param = problem.parameter("design_var")
        """
        _validate_constraint_inputs(name, boundary, "Parameter")

        param_var = variables_problem._create_static_parameter(
            self._multiphase_state.static_parameters, name, boundary
        )
        logger.debug("Static parameter created: name='%s'", name)
        return param_var

    def parameter_guess(self, **parameter_guesses: float) -> None:
        r"""Set initial guesses for static parameters.

        Args:
            \*\*parameter_guesses: Parameter names mapped to guess values

        Examples:
            >>> mass = problem.parameter("mass", boundary=(100, 500))
            >>> drag = problem.parameter("drag", boundary=(0, 0.1))
            >>> problem.parameter_guess(mass=300.0, drag=0.05)
        """
        if not parameter_guesses:
            return

        for name in parameter_guesses:
            if name not in self._multiphase_state.static_parameters.parameter_name_to_index:
                raise ConfigurationError(
                    f"Parameter '{name}' not defined. Call problem.parameter('{name}', ...) first.",
                    "Parameter guess configuration error",
                )

        self._multiphase_state.guess_static_parameters = parameter_guesses
        logger.debug("Static parameter guesses set: %s", list(parameter_guesses.keys()))

    def minimize(self, objective_expr: ca.MX | float | int) -> None:
        """
        Set objective function with comprehensive expression support.

        Defines scalar optimization objective supporting final states, integral
        terms, static parameters, and complex multi-phase expressions. Can combine
        multiple objective components with arbitrary mathematical relationships.

        Args:
            objective_expr: Scalar expression to minimize. Supports:

                - Final state values (state.final)
                - Integral terms (from add_integral())
                - Static parameters (from parameter())
                - Time variables (time.final)
                - Complex mathematical combinations

        Examples:
            Minimum time:

            >>> problem.minimize(t.final)

            Final state:

            >>> problem.minimize(-altitude.final)  # Maximize altitude

            Integral cost:

            >>> energy = phase.add_integral(thrust**2)
            >>> problem.minimize(energy)

            Multi-objective:

            >>> fuel_cost = phase.add_integral(thrust * price)
            >>> time_cost = t.final * rate
            >>> problem.minimize(fuel_cost + time_cost)

            Parameter optimization:

            >>> mass = problem.parameter("mass")
            >>> problem.minimize(mass - performance)
        """
        self._multiphase_state._functions_built = False
        variables_problem._set_multiphase_objective(self._multiphase_state, objective_expr)
        logger.info("Multiphase objective function defined")

    @property
    def _phases(self) -> dict[PhaseID, Any]:
        return self._multiphase_state.phases

    @property
    def _static_parameters(self) -> Any:
        return self._multiphase_state.static_parameters

    def _get_phase_ids(self) -> list[PhaseID]:
        return self._multiphase_state._get_phase_ids()

    def _get_phase_variable_counts(self, phase_id: PhaseID) -> tuple[int, int]:
        _validate_phase_exists(self._multiphase_state.phases, phase_id)
        return self._multiphase_state.phases[phase_id].get_variable_counts()

    def _get_total_variable_counts(self) -> tuple[int, int, int]:
        return self._multiphase_state._get_total_variable_counts()

    def _get_phase_ordered_state_names(self, phase_id: PhaseID) -> list[str]:
        _validate_phase_exists(self._multiphase_state.phases, phase_id)
        return self._multiphase_state.phases[phase_id].state_names.copy()

    def _get_phase_ordered_control_names(self, phase_id: PhaseID) -> list[str]:
        _validate_phase_exists(self._multiphase_state.phases, phase_id)
        return self._multiphase_state.phases[phase_id].control_names.copy()

    def _get_phase_dynamics_function(self, phase_id: PhaseID) -> Any:
        _validate_phase_exists(self._multiphase_state.phases, phase_id)
        phase_def = self._multiphase_state.phases[phase_id]

        if phase_def._dynamics_function is None:
            raise ValueError(
                f"Phase {phase_id} dynamics function not built - call validate_multiphase_configuration() first"
            )

        return phase_def._dynamics_function

    def _get_phase_numerical_dynamics_function(self, phase_id: PhaseID) -> Any:
        _validate_phase_exists(self._multiphase_state.phases, phase_id)
        phase_def = self._multiphase_state.phases[phase_id]

        if phase_def._numerical_dynamics_function is None:
            raise ValueError(
                f"Phase {phase_id} numerical dynamics function not built - call validate_multiphase_configuration() first"
            )

        return phase_def._numerical_dynamics_function

    def _get_objective_function(self) -> Any:
        if self._multiphase_state._objective_function is None:
            raise ValueError(
                "Multiphase objective function not built - call validate_multiphase_configuration() first"
            )

        return self._multiphase_state._objective_function

    def _get_phase_integrand_function(self, phase_id: PhaseID) -> Any:
        _validate_phase_exists(self._multiphase_state.phases, phase_id)
        phase_def = self._multiphase_state.phases[phase_id]
        return phase_def._integrand_function

    def _get_phase_path_constraints_function(self, phase_id: PhaseID) -> Any:
        _validate_phase_exists(self._multiphase_state.phases, phase_id)
        phase_def = self._multiphase_state.phases[phase_id]
        return phase_def._path_constraints_function

    def _get_cross_phase_event_constraints_function(self) -> Any:
        return self._multiphase_state._cross_phase_constraints_function

    def validate_multiphase_configuration(self) -> None:
        _process_symbolic_constraints_for_all_phases(
            self._multiphase_state.phases, self._multiphase_state.cross_phase_constraints
        )

        _validate_phase_requirements(self._multiphase_state.phases)

        if self._multiphase_state.objective_expression is None:
            raise ValueError("Problem must have objective function defined")

        _build_all_phase_functions(self._multiphase_state)

        logger.debug(
            "Multiphase configuration validated: %d phases, %d cross-phase constraints",
            len(self._multiphase_state.phases),
            len(self._multiphase_state.cross_phase_constraints),
        )



import numpy as np
import pytest
from numpy.testing import assert_allclose

from maptor.adaptive.phs.error_estimation import (
    _calculate_gamma_normalization_factors,
)
from maptor.adaptive.phs.initial_guess import (
    _interpolate_phase_trajectory_to_new_mesh_streamlined,
)
from maptor.adaptive.phs.numerical import (
    _map_global_normalized_tau_to_local_interval_tau,
    _map_local_interval_tau_to_global_normalized_tau,
    _map_local_tau_from_interval_k_plus_1_to_equivalent_in_interval_k,
    _map_local_tau_from_interval_k_to_equivalent_in_interval_k_plus_1,
)
from maptor.radau import _compute_radau_collocation_components


class TestCoordinateTransformationInvertibility:
    # Coordinate transformations must be mathematically exact."""

    @pytest.mark.parametrize(
        ("global_start", "global_end"),
        [
            (-1.0, 1.0),
            (-0.8, 0.3),
            (0.2, 0.9),
            (-0.1, 0.1),  # Small interval
        ],
    )
    def test_global_local_transformation_invertibility(self, global_start, global_end):
        test_points = np.linspace(global_start, global_end, 20)

        for global_tau in test_points:
            local_tau = _map_global_normalized_tau_to_local_interval_tau(
                global_tau, global_start, global_end
            )
            recovered_global = _map_local_interval_tau_to_global_normalized_tau(
                local_tau, global_start, global_end
            )

            assert abs(recovered_global - global_tau) < 1e-15

    @pytest.mark.parametrize(
        ("tau_start_k", "tau_shared", "tau_end_kp1"),
        [
            (-1.0, 0.0, 1.0),
            (-0.7, -0.2, 0.8),
            (-0.3, 0.1, 0.4),
        ],
    )
    def test_interval_mapping_composition_invertibility(self, tau_start_k, tau_shared, tau_end_kp1):
        test_points = np.linspace(-1, 1, 15)

        for local_tau_k in test_points:
            # Map from interval k to equivalent position in interval k+1
            mapped_tau_kp1 = _map_local_tau_from_interval_k_to_equivalent_in_interval_k_plus_1(
                local_tau_k, tau_start_k, tau_shared, tau_end_kp1
            )

            # Map back to interval k
            recovered_tau_k = _map_local_tau_from_interval_k_plus_1_to_equivalent_in_interval_k(
                mapped_tau_kp1, tau_start_k, tau_shared, tau_end_kp1
            )

            assert abs(recovered_tau_k - local_tau_k) < 1e-15


class TestInterpolationMathematicalExactness:
    @pytest.mark.parametrize(("N_source", "N_target"), [(3, 4), (4, 3), (3, 6), (5, 3)])
    def test_polynomial_interpolation_exactness(self, N_source, N_target):
        # Interpolating polynomial of degree ≤ min(N_source, N_target) must be exact.
        max_exact_degree = min(N_source, N_target)

        # Create source mesh and target mesh
        source_mesh = np.array([-1.0, 1.0])
        target_mesh = np.array([-1.0, 1.0])
        source_degrees = [N_source]
        target_degrees = [N_target]

        # Test polynomials up to the degree that should be exact
        for degree in range(max_exact_degree + 1):
            # Create polynomial trajectory on source mesh
            source_basis = _compute_radau_collocation_components(N_source)
            source_nodes = source_basis.state_approximation_nodes

            # Polynomial p(x) = x^degree
            poly_values = source_nodes**degree
            source_trajectory = [poly_values.reshape(1, -1)]

            # Interpolate to target mesh
            interpolated = _interpolate_phase_trajectory_to_new_mesh_streamlined(
                source_trajectory,
                source_mesh,
                source_degrees,
                target_mesh,
                target_degrees,
                num_variables=1,
                phase_id=1,
                is_state_trajectory=True,
            )

            # Evaluate exact polynomial at target nodes
            target_basis = _compute_radau_collocation_components(N_target)
            target_nodes = target_basis.state_approximation_nodes
            exact_values = target_nodes**degree

            # Interpolation should be exact for polynomials within degree bounds
            interpolated_values = interpolated[0].flatten()
            max_error = np.max(np.abs(interpolated_values - exact_values))

            assert max_error < 1e-12, f"Polynomial degree {degree} not exact: error = {max_error}"

    def test_interpolation_preserves_constant_function(self):
        constant_value = 3.14159

        # Various mesh configurations
        test_cases = [
            ([3], [5], np.array([-1.0, 1.0])),
            ([4, 3], [2, 6], np.array([-1.0, 0.0, 1.0])),
        ]

        for source_degrees, target_degrees, mesh_points in test_cases:
            # Create constant trajectory
            source_trajectories = []
            for N in source_degrees:
                basis = _compute_radau_collocation_components(N)
                num_nodes = len(basis.state_approximation_nodes)
                constant_traj = np.full((1, num_nodes), constant_value)
                source_trajectories.append(constant_traj)

            # Interpolate
            interpolated = _interpolate_phase_trajectory_to_new_mesh_streamlined(
                source_trajectories,
                mesh_points,
                source_degrees,
                mesh_points,
                target_degrees,
                num_variables=1,
                phase_id=1,
                is_state_trajectory=True,
            )

            # All interpolated values should equal the constant
            for traj in interpolated:
                assert_allclose(traj, constant_value, rtol=1e-15)


class TestErrorEstimationConsistency:
    def test_gamma_normalization_bounds(self):
        test_cases = [
            np.array([0.1, 1.0, 10.0]),  # Normal values
            np.array([1e-15, 1e-10, 1e15]),  # Extreme values
            np.array([0.0, 0.0, 0.0]),  # Zero values
            np.array([1e20, 1e20, 1e20]),  # Large values
        ]

        for max_state_values in test_cases:
            gamma_factors = _calculate_gamma_normalization_factors(max_state_values)

            # Must be finite and positive
            assert np.all(np.isfinite(gamma_factors))
            assert np.all(gamma_factors > 0)

            # Must match actual implementation: 1/max(1+max_val, 1e-12)
            gamma_denominator = 1.0 + max_state_values
            safe_denominator = np.maximum(gamma_denominator, 1e-12)
            expected = (1.0 / safe_denominator).reshape(-1, 1)

            assert_allclose(gamma_factors, expected, rtol=1e-15)

    def test_gamma_normalization_monotonicity(self):
        small_values = np.array([0.1, 0.5])
        large_values = np.array([10.0, 50.0])

        gamma_small = _calculate_gamma_normalization_factors(small_values)
        gamma_large = _calculate_gamma_normalization_factors(large_values)

        # Gamma factors should decrease as state values increase
        assert np.all(gamma_small > gamma_large)


class TestMeshRefinementConsistency:
    @pytest.mark.parametrize("original_degree", [3, 4, 5])
    def test_p_refinement_preserves_polynomials(self, original_degree):
        refined_degree = original_degree + 2

        # Create polynomial trajectory on original mesh
        original_basis = _compute_radau_collocation_components(original_degree)
        original_nodes = original_basis.state_approximation_nodes

        # Test polynomial of degree = original_degree (should be preserved exactly)
        test_polynomial_degree = original_degree
        poly_values = original_nodes**test_polynomial_degree
        original_trajectory = [poly_values.reshape(1, -1)]

        # "Refine" to higher degree mesh (simulate p-refinement)
        mesh_points = np.array([-1.0, 1.0])
        original_degrees = [original_degree]
        refined_degrees = [refined_degree]

        refined_trajectory = _interpolate_phase_trajectory_to_new_mesh_streamlined(
            original_trajectory,
            mesh_points,
            original_degrees,
            mesh_points,
            refined_degrees,
            num_variables=1,
            phase_id=1,
            is_state_trajectory=True,
        )

        # Evaluate exact polynomial at refined nodes
        refined_basis = _compute_radau_collocation_components(refined_degree)
        refined_nodes = refined_basis.state_approximation_nodes
        exact_refined_values = refined_nodes**test_polynomial_degree

        # Should be exact since polynomial degree ≤ original degree
        refined_values = refined_trajectory[0].flatten()
        max_error = np.max(np.abs(refined_values - exact_refined_values))

        assert max_error < 1e-11, f"P-refinement failed to preserve polynomial: error = {max_error}"

    def test_h_refinement_preserves_total_integral(self):
        # Create a function on coarse mesh
        coarse_degree = 4
        coarse_basis = _compute_radau_collocation_components(coarse_degree)
        coarse_nodes = coarse_basis.state_approximation_nodes

        # Simple quadratic function: f(x) = x^2 + 1
        func_values = coarse_nodes**2 + 1
        coarse_trajectory = [func_values.reshape(1, -1)]

        # Compute integral on coarse mesh using quadrature
        coarse_weights = coarse_basis.quadrature_weights
        coarse_colloc_nodes = coarse_basis.collocation_nodes
        coarse_func_at_colloc = coarse_colloc_nodes**2 + 1
        coarse_integral = np.sum(coarse_weights * coarse_func_at_colloc)

        # Refine mesh (h-refinement): split into two intervals
        coarse_mesh = np.array([-1.0, 1.0])
        refined_mesh = np.array([-1.0, 0.0, 1.0])
        refined_degrees = [coarse_degree, coarse_degree]

        refined_trajectories = _interpolate_phase_trajectory_to_new_mesh_streamlined(
            coarse_trajectory,
            coarse_mesh,
            [coarse_degree],
            refined_mesh,
            refined_degrees,
            num_variables=1,
            phase_id=1,
            is_state_trajectory=True,
        )

        # Compute integral on refined mesh
        refined_integral = 0.0
        for k, _traj in enumerate(refined_trajectories):
            basis = _compute_radau_collocation_components(refined_degrees[k])
            weights = basis.quadrature_weights
            colloc_nodes = basis.collocation_nodes

            # Transform collocation nodes to global interval
            if k == 0:  # First interval: [-1, 0]
                global_colloc = (colloc_nodes - 1) / 2  # Map [-1,1] to [-1,0]
                scaling = 0.5
            else:  # Second interval: [0, 1]
                global_colloc = (colloc_nodes + 1) / 2  # Map [-1,1] to [0,1]
                scaling = 0.5

            # Evaluate function at global collocation points
            func_at_global_colloc = global_colloc**2 + 1
            refined_integral += scaling * np.sum(weights * func_at_global_colloc)

        # Integrals should match (h-refinement preserves integrals for smooth functions)
        integral_error = abs(refined_integral - coarse_integral)
        assert integral_error < 1e-10, f"H-refinement changed integral: error = {integral_error}"



import casadi as ca
import numpy as np
import pytest

import maptor as mtor
from maptor.direct_solver.initial_guess_solver import _apply_multiphase_initial_guess
from maptor.direct_solver.variables_solver import (
    _setup_multiphase_optimization_variables,
    setup_phase_interval_state_variables,
)
from maptor.exceptions import ConfigurationError
from maptor.input_validation import _validate_multiphase_problem_ready_for_solving


class TestVariableSetupIntegration:
    def test_single_phase_variable_creation(self):
        problem = mtor.Problem("Single Phase Setup")
        phase = problem.set_phase(1)

        t = phase.time(initial=0, final=1)
        x = phase.state("position", initial=0, final=1)
        v = phase.state("velocity", initial=0)
        u = phase.control("thrust", boundary=(0, 100))

        phase.dynamics({x: v, v: u})
        problem.minimize(t.final)

        # Configure mesh
        phase.mesh([3, 4], [-1, 0, 1])

        # Create solver and test variable setup
        opti = ca.Opti()
        variables = _setup_multiphase_optimization_variables(opti, problem)  # type: ignore[arg-type]

        # Verify phase variables structure
        assert 1 in variables.phase_variables
        phase_vars = variables.phase_variables[1]

        # Check time variables
        assert phase_vars.initial_time is not None
        assert phase_vars.terminal_time is not None
        assert isinstance(phase_vars.initial_time, ca.MX)
        assert isinstance(phase_vars.terminal_time, ca.MX)

        expected_nodes = 2 + 1  # num_intervals + 1
        assert len(phase_vars.state_at_mesh_nodes) == expected_nodes

        # Check control variables (one per interval)
        assert len(phase_vars.control_variables) == 2  # num_intervals

        # Check dimensions match problem definition
        num_states, num_controls = problem._get_phase_variable_counts(1)
        assert num_states == 2  # position, velocity
        assert num_controls == 1  # thrust

    def test_multiphase_variable_organization(self):
        problem = mtor.Problem("Multiphase Setup")

        # Phase 1: 2 states, 1 control
        p1 = problem.set_phase(1)
        _t1 = p1.time(initial=0, final=1)
        x1 = p1.state("x", initial=0)
        v1 = p1.state("v", initial=0)
        u1 = p1.control("u")
        p1.dynamics({x1: v1, v1: u1})
        p1.mesh([3], [-1, 1])

        # Phase 2: 3 states, 2 controls
        p2 = problem.set_phase(2)
        t2 = p2.time(initial=1, final=2)
        x2 = p2.state("x", initial=x1.final)
        y2 = p2.state("y", initial=0)
        v2 = p2.state("v", initial=v1.final)
        u2 = p2.control("u")
        w2 = p2.control("w")
        p2.dynamics({x2: v2, y2: w2, v2: u2})
        p2.mesh([4, 4], [-1, 0, 1])

        problem.minimize(t2.final)

        opti = ca.Opti()
        variables = _setup_multiphase_optimization_variables(opti, problem)  # type: ignore[arg-type]

        # Verify both phases are present
        assert 1 in variables.phase_variables
        assert 2 in variables.phase_variables

        # Check phase 1 structure
        p1_vars = variables.phase_variables[1]
        assert len(p1_vars.state_at_mesh_nodes) == 2  # 1 interval + 1
        assert len(p1_vars.control_variables) == 1  # 1 interval

        # Check phase 2 structure
        p2_vars = variables.phase_variables[2]
        assert len(p2_vars.state_at_mesh_nodes) == 3  # 2 intervals + 1
        assert len(p2_vars.control_variables) == 2  # 2 intervals

    def test_static_parameter_integration(self):
        problem = mtor.Problem("Parameter Integration")

        # Create static parameters
        mass = problem.parameter("mass", boundary=(100, 1000))
        thrust_max = problem.parameter("thrust_max", boundary=(500, 2000))

        phase = problem.set_phase(1)
        t = phase.time(initial=0, final=1)
        v = phase.state("velocity", initial=0)
        u = phase.control("throttle", boundary=(0, 1))

        # Use parameters in dynamics
        phase.dynamics({v: u * thrust_max / mass})
        problem.minimize(t.final)

        phase.mesh([3], [-1, 1])

        opti = ca.Opti()
        variables = _setup_multiphase_optimization_variables(opti, problem)  # type: ignore[arg-type]

        # Verify static parameters were created
        assert variables.static_parameters is not None
        assert isinstance(variables.static_parameters, ca.MX)

        # Check parameter count matches
        _, _, num_static_params = problem._get_total_variable_counts()
        assert num_static_params == 2

    def test_integral_variable_creation(self):
        problem = mtor.Problem("Integral Variables")
        phase = problem.set_phase(1)

        _t = phase.time(initial=0, final=1)
        x = phase.state("x", initial=0)
        u = phase.control("u")

        phase.dynamics({x: u})

        # Add multiple integrals
        cost1 = phase.add_integral(x**2)
        cost2 = phase.add_integral(u**2)
        cost3 = phase.add_integral(1.0)  # Time integral

        problem.minimize(cost1 + 0.1 * cost2 + 0.01 * cost3)

        phase.mesh([3], [-1, 1])

        opti = ca.Opti()
        variables = _setup_multiphase_optimization_variables(opti, problem)  # type: ignore[arg-type]

        phase_vars = variables.phase_variables[1]

        # Verify integral variables were created
        assert phase_vars.integral_variables is not None
        assert isinstance(phase_vars.integral_variables, ca.MX)

        # Check integral count
        assert phase._phase_def.num_integrals == 3


class TestMeshIntervalHandling:
    def test_uniform_mesh_interval_setup(self):
        problem = mtor.Problem("Uniform Mesh")
        phase = problem.set_phase(1)

        x = phase.state("x", initial=0, final=1)
        v = phase.state("v", initial=0)
        u = phase.control("u")

        phase.dynamics({x: v, v: u})
        problem.minimize(x.final)

        # Uniform mesh: 3 intervals, each with 4 collocation points
        phase.mesh([4, 4, 4], [-1, -0.3, 0.3, 1])

        opti = ca.Opti()
        variables = _setup_multiphase_optimization_variables(opti, problem)  # type: ignore[arg-type]

        phase_vars = variables.phase_variables[1]

        # Test individual interval setup
        for k in range(3):  # 3 intervals
            state_matrix, interior_nodes = setup_phase_interval_state_variables(
                opti, 1, k, 2, 4, phase_vars.state_at_mesh_nodes
            )

            # Each interval should have proper state matrix structure
            assert state_matrix is not None
            assert isinstance(state_matrix, ca.MX)

            # Interior nodes for degree 4 should exist (degree > 1)
            assert interior_nodes is not None

    def test_non_uniform_mesh_interval_setup(self):
        problem = mtor.Problem("Non-uniform Mesh")
        phase = problem.set_phase(1)

        x = phase.state("x", initial=0)
        u = phase.control("u")

        phase.dynamics({x: u})
        problem.minimize(x.final)

        # Non-uniform mesh: different degrees per interval
        phase.mesh([3, 6, 4], [-1, -0.5, 0.5, 1])

        opti = ca.Opti()
        variables = _setup_multiphase_optimization_variables(opti, problem)  # type: ignore[arg-type]

        phase_vars = variables.phase_variables[1]

        # Control variables should match mesh structure
        assert len(phase_vars.control_variables) == 3  # 3 intervals

        # Each control variable should have correct dimensions
        expected_colloc_points = [3, 6, 4]
        for k, expected_points in enumerate(expected_colloc_points):
            control_var = phase_vars.control_variables[k]
            assert control_var.shape[1] == expected_points  # num_controls x num_colloc_points

    def test_minimal_mesh_setup(self):
        problem = mtor.Problem("Minimal Mesh")
        phase = problem.set_phase(1)

        x = phase.state("x", initial=0, final=1)
        u = phase.control("u")

        phase.dynamics({x: u})
        problem.minimize(x.final)

        # Minimal mesh: single interval
        phase.mesh([3], [-1, 1])

        opti = ca.Opti()
        variables = _setup_multiphase_optimization_variables(opti, problem)  # type: ignore[arg-type]

        phase_vars = variables.phase_variables[1]

        # Single interval setup
        assert len(phase_vars.state_at_mesh_nodes) == 2  # 1 interval + 1
        assert len(phase_vars.control_variables) == 1  # 1 interval

    def test_high_order_mesh_setup(self):
        problem = mtor.Problem("High Order Mesh")
        phase = problem.set_phase(1)

        x = phase.state("x", initial=0)
        v = phase.state("v", initial=0)
        u = phase.control("u")

        phase.dynamics({x: v, v: u})
        problem.minimize(x.final)

        # High-order mesh
        phase.mesh([10, 12], [-1, 0, 1])

        opti = ca.Opti()
        variables = _setup_multiphase_optimization_variables(opti, problem)  # type: ignore[arg-type]

        phase_vars = variables.phase_variables[1]

        # High-order intervals should create appropriate structures
        assert len(phase_vars.control_variables) == 2  # 2 intervals

        # Check control variable dimensions
        assert phase_vars.control_variables[0].shape[1] == 10  # First interval: 10 points
        assert phase_vars.control_variables[1].shape[1] == 12  # Second interval: 12 points


class TestInitialGuessApplication:
    def test_complete_initial_guess_application(self):
        problem = mtor.Problem("Complete Guess")
        phase = problem.set_phase(1)

        _t = phase.time(initial=0, final=1)
        x = phase.state("position", initial=0, final=1)
        v = phase.state("velocity", initial=0)
        u = phase.control("thrust")

        phase.dynamics({x: v, v: u})

        energy = phase.add_integral(u**2)
        problem.minimize(energy)

        phase.mesh([3, 4], [-1, 0, 1])

        states_guess = [
            np.array(
                [
                    [0.0, 0.25, 0.5, 0.75],
                    [0.0, 0.5, 1.0, 1.0],
                ]
            ),
            np.array(
                [
                    [0.75, 0.8, 0.9, 0.95, 1.0],
                    [1.0, 0.8, 0.6, 0.3, 0.0],
                ]
            ),
        ]

        controls_guess = [
            np.array([[1.0, 1.0, 1.0]]),
            np.array([[-1.0, -1.0, -1.0, -1.0]]),
        ]

        phase.guess(
            states=states_guess,
            controls=controls_guess,
            initial_time=0.0,
            terminal_time=2.0,
            integrals=1.5,
        )

        opti = ca.Opti()
        variables = _setup_multiphase_optimization_variables(opti, problem)  # type: ignore[arg-type]

        # Manually populate interior_variables to match mesh structure (2 intervals)
        variables.phase_variables[1].interior_variables = [None, None]

        _apply_multiphase_initial_guess(opti, variables, problem)  # type: ignore[arg-type]

    def test_partial_initial_guess_application(self):
        problem = mtor.Problem("Partial Guess")
        phase = problem.set_phase(1)

        t = phase.time(initial=0)  # Free final time
        x = phase.state("x", initial=0)
        u = phase.control("u")

        phase.dynamics({x: u})
        problem.minimize(t.final)

        phase.mesh([3], [-1, 1])

        # Only provide terminal time guess (partial)
        phase.guess(terminal_time=5.0)

        opti = ca.Opti()
        variables = _setup_multiphase_optimization_variables(opti, problem)  # type: ignore[arg-type]

        # Should handle partial guess without errors
        _apply_multiphase_initial_guess(opti, variables, problem)  # type: ignore[arg-type]


class TestValidationIntegration:
    def test_problem_validation_before_setup(self):
        problem = mtor.Problem("Validation Test")
        phase = problem.set_phase(1)

        x = phase.state("x", initial=0)
        _u = phase.control("u")

        # Missing dynamics - should be caught by validation
        problem.minimize(x.final)
        phase.mesh([3], [-1, 1])

        with pytest.raises(
            ConfigurationError, match="Phase 1 dynamics must be defined - call phase.dynamics()"
        ):
            _validate_multiphase_problem_ready_for_solving(problem)  # type: ignore[arg-type]

    def test_mesh_configuration_validation(self):
        problem = mtor.Problem("Mesh Validation")
        phase = problem.set_phase(1)

        x = phase.state("x", initial=0)
        u = phase.control("u")

        phase.dynamics({x: u})
        problem.minimize(x.final)

        # Missing mesh configuration
        with pytest.raises(ConfigurationError, match="mesh must be configured"):
            _validate_multiphase_problem_ready_for_solving(problem)  # type: ignore[arg-type]

    def test_objective_validation(self):
        problem = mtor.Problem("Objective Validation")
        phase = problem.set_phase(1)

        x = phase.state("x", initial=0)
        u = phase.control("u")

        phase.dynamics({x: u})
        phase.mesh([3], [-1, 1])

        # Missing objective
        with pytest.raises(ConfigurationError, match="must have objective"):
            _validate_multiphase_problem_ready_for_solving(problem)  # type: ignore[arg-type]


class TestBoundaryConstraintApplication:
    def test_time_bound_constraints(self):
        problem = mtor.Problem("Time Bounds")
        phase = problem.set_phase(1)

        # Bounded initial and final times
        t = phase.time(initial=(0, 1), final=(5, 10))
        x = phase.state("x", initial=0)
        u = phase.control("u")

        phase.dynamics({x: u})
        problem.minimize(t.final)

        phase.mesh([3], [-1, 1])

        opti = ca.Opti()
        variables = _setup_multiphase_optimization_variables(opti, problem)  # type: ignore[arg-type]

        # Time bounds should be applied as constraints
        # Verify by checking that opti has constraints (hard to verify specific constraints)
        assert variables.phase_variables[1].initial_time is not None
        assert variables.phase_variables[1].terminal_time is not None

    def test_state_bound_constraints(self):
        problem = mtor.Problem("State Bounds")
        phase = problem.set_phase(1)

        t = phase.time(initial=0, final=1)

        # States with various boundary constraints
        pos = phase.state("position", initial=0, final=1, boundary=(0, 10))
        vel = phase.state("velocity", initial=0, boundary=(-5, 5))

        u = phase.control("thrust", boundary=(-2, 2))

        phase.dynamics({pos: vel, vel: u})
        problem.minimize(t.final)

        phase.mesh([3], [-1, 1])

        opti = ca.Opti()
        variables = _setup_multiphase_optimization_variables(opti, problem)  # type: ignore[arg-type]

        # Boundary constraints should be set up for path constraint processing
        phase_vars = variables.phase_variables[1]
        assert len(phase_vars.state_at_mesh_nodes) == 2  # 1 interval + 1
        assert len(phase_vars.control_variables) == 1  # 1 interval

    def test_control_bound_constraints(self):
        problem = mtor.Problem("Control Bounds")
        phase = problem.set_phase(1)

        x = phase.state("x", initial=0)

        # Controls with different bound types
        thrust = phase.control("thrust", boundary=(0, 100))  # Two-sided bounds
        steering = phase.control("steering", boundary=(-30, 30))  # Symmetric bounds
        fixed_ctrl = phase.control("fixed", boundary=5.0)  # Fixed control

        phase.dynamics({x: thrust + steering + fixed_ctrl})
        problem.minimize(x.final)

        phase.mesh([3], [-1, 1])

        opti = ca.Opti()
        variables = _setup_multiphase_optimization_variables(opti, problem)  # type: ignore[arg-type]

        # Control variables should be created properly
        phase_vars = variables.phase_variables[1]
        control_var = phase_vars.control_variables[0]

        # Should have 3 controls and 3 collocation points
        assert control_var.shape[0] == 3  # num_controls
        assert control_var.shape[1] == 3  # num_collocation_points



import logging
from dataclasses import dataclass
from typing import Any

import casadi as ca
import numpy as np

from ..input_validation import _set_integral_guess_values
from ..mtor_types import FloatArray, PhaseID, ProblemProtocol
from .types_solver import _MultiPhaseVariable, _PhaseVariable


logger = logging.getLogger(__name__)


@dataclass
class _PhaseGuessContext:
    opti: ca.Opti
    phase_vars: _PhaseVariable
    phase_id: PhaseID
    num_states: int
    num_controls: int
    num_mesh_intervals: int
    num_integrals: int
    initial_guess: Any


def _apply_state_guesses(context: _PhaseGuessContext, phase_states: list[FloatArray]) -> None:
    # Apply global mesh node states
    for k in range(context.num_mesh_intervals):
        state_guess_k = phase_states[k]
        num_provided_states = state_guess_k.shape[0]

        # Set initial node guess (only once)
        if k == 0:
            initial_var = context.phase_vars.state_at_mesh_nodes[0]
            if num_provided_states == context.num_states:
                context.opti.set_initial(initial_var, state_guess_k[:, 0])
            else:
                logger.info(
                    f"Phase {context.phase_id} initial state guess has {num_provided_states} states, "
                    f"expected {context.num_states}. Using CasADi defaults for missing states."
                )

        # Set terminal node guess
        terminal_var = context.phase_vars.state_at_mesh_nodes[k + 1]
        if num_provided_states == context.num_states:
            context.opti.set_initial(terminal_var, state_guess_k[:, -1])
        else:
            logger.info(
                f"Phase {context.phase_id} terminal state guess has {num_provided_states} states, "
                f"expected {context.num_states}. Using CasADi defaults for missing states."
            )

    # Apply interior state node guesses
    for k in range(context.num_mesh_intervals):
        interior_var = context.phase_vars.interior_variables[k]
        if interior_var is not None:
            state_guess_k = phase_states[k]
            num_provided_states = state_guess_k.shape[0]

            if num_provided_states == context.num_states:
                num_interior_nodes = interior_var.shape[1]
                interior_guess = state_guess_k[:, 1 : 1 + num_interior_nodes]
                context.opti.set_initial(interior_var, interior_guess)
            else:
                logger.info(
                    f"Phase {context.phase_id} interior state guess has {num_provided_states} states, "
                    f"expected {context.num_states}. Using CasADi defaults for missing states."
                )


def _apply_control_guesses(context: _PhaseGuessContext, phase_controls: list[FloatArray]) -> None:
    for k in range(context.num_mesh_intervals):
        control_guess_k = phase_controls[k]
        num_provided_controls = control_guess_k.shape[0]

        if num_provided_controls == context.num_controls:
            context.opti.set_initial(context.phase_vars.control_variables[k], control_guess_k)
        else:
            logger.info(
                f"Phase {context.phase_id} control guess has {num_provided_controls} controls, "
                f"expected {context.num_controls}. Using CasADi defaults for missing controls."
            )


def _apply_integral_guesses(
    context: _PhaseGuessContext, phase_integrals: float | FloatArray
) -> None:
    if context.num_integrals > 0 and context.phase_vars.integral_variables is not None:
        _set_integral_guess_values(
            context.opti,
            context.phase_vars.integral_variables,
            phase_integrals,
            context.num_integrals,
        )


def _apply_phase_guesses_from_phase_definition(
    opti: ca.Opti,
    phase_vars: _PhaseVariable,
    phase_def: Any,  # PhaseDefinition
    problem: ProblemProtocol,
    phase_id: PhaseID,
) -> None:
    num_states, num_controls = problem._get_phase_variable_counts(phase_id)
    num_mesh_intervals = len(phase_def.collocation_points_per_interval)
    num_integrals = phase_def.num_integrals

    # Apply time guesses using existing logic
    if phase_def.guess_initial_time is not None:
        opti.set_initial(phase_vars.initial_time, phase_def.guess_initial_time)

    if phase_def.guess_terminal_time is not None:
        opti.set_initial(phase_vars.terminal_time, phase_def.guess_terminal_time)

    # Apply state guesses using existing logic
    if phase_def.guess_states is not None:
        context = _PhaseGuessContext(
            opti=opti,
            phase_vars=phase_vars,
            phase_id=phase_id,
            num_states=num_states,
            num_controls=num_controls,
            num_mesh_intervals=num_mesh_intervals,
            num_integrals=num_integrals,
            initial_guess=None,
        )
        _apply_state_guesses(context, phase_def.guess_states)

    # Apply control guesses using existing logic
    if phase_def.guess_controls is not None:
        context = _PhaseGuessContext(
            opti=opti,
            phase_vars=phase_vars,
            phase_id=phase_id,
            num_states=num_states,
            num_controls=num_controls,
            num_mesh_intervals=num_mesh_intervals,
            num_integrals=num_integrals,
            initial_guess=None,
        )
        _apply_control_guesses(context, phase_def.guess_controls)

    # Apply integral guesses using existing logic
    if phase_def.guess_integrals is not None:
        context = _PhaseGuessContext(
            opti=opti,
            phase_vars=phase_vars,
            phase_id=phase_id,
            num_states=num_states,
            num_controls=num_controls,
            num_mesh_intervals=num_mesh_intervals,
            num_integrals=num_integrals,
            initial_guess=None,
        )
        _apply_integral_guesses(context, phase_def.guess_integrals)


def _apply_multiphase_initial_guess(
    opti: ca.Opti,
    variables: _MultiPhaseVariable,
    problem: ProblemProtocol,
) -> None:
    # Apply phase-level guesses
    for phase_id in problem._get_phase_ids():
        if phase_id in variables.phase_variables:
            phase_vars = variables.phase_variables[phase_id]
            phase_def = problem._phases[phase_id]
            _apply_phase_guesses_from_phase_definition(
                opti, phase_vars, phase_def, problem, phase_id
            )

    # Apply static parameter guesses
    if (
        problem._multiphase_state.guess_static_parameters is not None
        and variables.static_parameters is not None
    ):
        param_names = problem._static_parameters.parameter_names
        param_guesses_dict = problem._multiphase_state.guess_static_parameters

        ordered_guesses = []
        for name in param_names:
            if name in param_guesses_dict:
                ordered_guesses.append(param_guesses_dict[name])
            else:
                ordered_guesses.append(0.0)

        opti.set_initial(variables.static_parameters, np.array(ordered_guesses, dtype=np.float64))



import functools
from dataclasses import dataclass, field
from typing import Literal, cast, overload

import numpy as np
from scipy.special import roots_jacobi as _scipy_roots_jacobi

from .input_validation import _validate_positive_integer
from .mtor_types import FloatArray
from .utils.constants import NUMERICAL_ZERO


@dataclass
class RadauBasisComponents:
    """Components for Radau pseudospectral method basis functions."""

    state_approximation_nodes: FloatArray = field(
        default_factory=lambda: np.array([], dtype=np.float64)
    )
    collocation_nodes: FloatArray = field(default_factory=lambda: np.array([], dtype=np.float64))
    quadrature_weights: FloatArray = field(default_factory=lambda: np.array([], dtype=np.float64))
    differentiation_matrix: FloatArray = field(
        default_factory=lambda: np.empty((0, 0), dtype=np.float64)
    )
    barycentric_weights_for_state_nodes: FloatArray = field(
        default_factory=lambda: np.array([], dtype=np.float64)
    )
    barycentric_weights_for_collocation_nodes: FloatArray = field(
        default_factory=lambda: np.array([], dtype=np.float64)
    )
    lagrange_at_tau_plus_one: FloatArray = field(
        default_factory=lambda: np.array([], dtype=np.float64)
    )


@dataclass
class RadauNodesAndWeights:
    state_approximation_nodes: FloatArray
    collocation_nodes: FloatArray
    quadrature_weights: FloatArray


@overload
def _roots_jacobi(
    n: int, alpha: float, beta: float, mu: Literal[False]
) -> tuple[FloatArray, FloatArray]: ...


@overload
def _roots_jacobi(
    n: int, alpha: float, beta: float, mu: Literal[True]
) -> tuple[FloatArray, FloatArray, float]: ...


def _roots_jacobi(
    n: int, alpha: float, beta: float, mu: bool = False
) -> tuple[FloatArray, FloatArray] | tuple[FloatArray, FloatArray, float]:
    if mu:
        result = _scipy_roots_jacobi(n, alpha, beta, mu=True)
        x_val = result[0]
        w_val = result[1]
        mu_val: float = result[2]
        return (
            x_val.astype(np.float64),
            w_val.astype(np.float64),
            float(mu_val),
        )
    else:
        result = _scipy_roots_jacobi(n, alpha, beta, mu=False)
        x_val = result[0]
        w_val = result[1]
        return (
            x_val.astype(np.float64),
            w_val.astype(np.float64),
        )


def _compute_legendre_gauss_radau_nodes_and_weights(
    num_collocation_nodes: int,
) -> RadauNodesAndWeights:
    collocation_nodes_list: list[float] = [-1.0]

    if num_collocation_nodes == 1:
        quadrature_weights_list: list[float] = [2.0]
    else:
        num_interior_roots = num_collocation_nodes - 1
        interior_roots, jacobi_weights, _ = _roots_jacobi(num_interior_roots, 0.0, 1.0, mu=True)
        interior_weights = jacobi_weights / (np.add(1.0, interior_roots))
        left_endpoint_weight = 2.0 / (num_collocation_nodes**2)
        collocation_nodes_list.extend(list(interior_roots))
        quadrature_weights_list = [left_endpoint_weight, *list(interior_weights)]

    final_collocation_nodes = np.array(collocation_nodes_list, dtype=np.float64)
    final_quadrature_weights = np.array(quadrature_weights_list, dtype=np.float64)

    state_approximation_nodes_temp = np.concatenate(
        [final_collocation_nodes, np.array([1.0], dtype=np.float64)]
    )
    state_approximation_nodes = np.unique(state_approximation_nodes_temp)

    return RadauNodesAndWeights(
        state_approximation_nodes=state_approximation_nodes,
        collocation_nodes=final_collocation_nodes,
        quadrature_weights=final_quadrature_weights,
    )


def _compute_barycentric_weights(nodes: FloatArray) -> FloatArray:
    num_nodes = len(nodes)
    if num_nodes == 1:
        return np.array([1.0], dtype=np.float64)

    nodes_col = nodes[:, np.newaxis]
    nodes_row = nodes[np.newaxis, :]
    differences_matrix = nodes_col - nodes_row

    diagonal_mask = np.eye(num_nodes, dtype=bool)

    near_zero_mask = np.abs(differences_matrix) < NUMERICAL_ZERO
    perturbation = np.sign(differences_matrix) * NUMERICAL_ZERO
    perturbation[perturbation == 0] = NUMERICAL_ZERO

    off_diagonal_near_zero = near_zero_mask & ~diagonal_mask
    differences_matrix = np.where(off_diagonal_near_zero, perturbation, differences_matrix)

    differences_matrix[diagonal_mask] = 1.0

    products = np.prod(differences_matrix, axis=1, dtype=np.float64)

    small_product_mask = np.abs(products) < NUMERICAL_ZERO**2
    safe_products = np.where(
        small_product_mask,
        np.where(products == 0, 1.0 / (NUMERICAL_ZERO**2), np.sign(products) / (NUMERICAL_ZERO**2)),
        1.0 / products,
    )

    return safe_products.astype(np.float64)


def _evaluate_lagrange_polynomial_at_point(
    polynomial_definition_nodes: FloatArray,
    barycentric_weights: FloatArray,
    evaluation_point_tau: float,
) -> FloatArray:
    num_nodes = len(polynomial_definition_nodes)

    differences = np.abs(evaluation_point_tau - polynomial_definition_nodes)
    coincident_mask = differences < NUMERICAL_ZERO

    if np.any(coincident_mask):
        lagrange_values = np.zeros(num_nodes, dtype=np.float64)
        lagrange_values[coincident_mask] = 1.0
        return lagrange_values

    diffs = evaluation_point_tau - polynomial_definition_nodes

    near_zero_mask = np.abs(diffs) < NUMERICAL_ZERO
    safe_diffs = np.where(
        near_zero_mask, np.where(diffs == 0, NUMERICAL_ZERO, np.sign(diffs) * NUMERICAL_ZERO), diffs
    )

    terms = barycentric_weights / safe_diffs
    sum_terms = np.sum(terms)

    if abs(sum_terms) < NUMERICAL_ZERO:
        return np.zeros(num_nodes, dtype=np.float64)

    normalized_terms = terms / sum_terms
    return cast(FloatArray, normalized_terms)


def _compute_lagrange_derivative_coefficients_at_point(
    polynomial_definition_nodes: FloatArray,
    barycentric_weights: FloatArray,
    evaluation_point_tau: float,
) -> FloatArray:
    num_nodes = len(polynomial_definition_nodes)

    differences = np.abs(evaluation_point_tau - polynomial_definition_nodes)
    matched_indices = np.where(differences < NUMERICAL_ZERO)[0]

    if len(matched_indices) == 0:
        return np.zeros(num_nodes, dtype=np.float64)

    matched_node_idx_k = matched_indices[0]
    derivatives = np.zeros(num_nodes, dtype=np.float64)

    node_diffs = polynomial_definition_nodes[matched_node_idx_k] - polynomial_definition_nodes

    near_zero_mask = np.abs(node_diffs) < NUMERICAL_ZERO
    safe_diffs = np.where(
        near_zero_mask,
        np.where(node_diffs == 0, NUMERICAL_ZERO, np.sign(node_diffs) * NUMERICAL_ZERO),
        node_diffs,
    )

    non_diagonal_mask = np.arange(num_nodes) != matched_node_idx_k

    if abs(barycentric_weights[matched_node_idx_k]) < NUMERICAL_ZERO:
        derivatives[non_diagonal_mask] = 0.0
    else:
        weight_ratios = barycentric_weights / barycentric_weights[matched_node_idx_k]
        derivatives[non_diagonal_mask] = (
            weight_ratios[non_diagonal_mask] / safe_diffs[non_diagonal_mask]
        )

    derivatives[matched_node_idx_k] = np.sum(1.0 / safe_diffs[non_diagonal_mask])

    return derivatives


@functools.lru_cache(maxsize=32)
def _compute_radau_collocation_components(
    num_collocation_nodes: int,
) -> RadauBasisComponents:
    _validate_positive_integer(num_collocation_nodes, "collocation nodes")

    lgr_data = _compute_legendre_gauss_radau_nodes_and_weights(num_collocation_nodes)

    state_nodes = lgr_data.state_approximation_nodes
    collocation_nodes = lgr_data.collocation_nodes
    quadrature_weights = lgr_data.quadrature_weights

    num_state_nodes = len(state_nodes)
    num_actual_collocation_nodes = len(collocation_nodes)

    bary_weights_state_nodes = _compute_barycentric_weights(state_nodes)
    bary_weights_collocation_nodes = _compute_barycentric_weights(collocation_nodes)

    diff_matrix = np.zeros((num_actual_collocation_nodes, num_state_nodes), dtype=np.float64)
    for i in range(num_actual_collocation_nodes):
        tau_c_i = collocation_nodes[i]
        diff_matrix[i, :] = _compute_lagrange_derivative_coefficients_at_point(
            state_nodes, bary_weights_state_nodes, tau_c_i
        )

    lagrange_at_tau_plus_one = _evaluate_lagrange_polynomial_at_point(
        state_nodes, bary_weights_state_nodes, 1.0
    )

    return RadauBasisComponents(
        state_approximation_nodes=state_nodes,
        collocation_nodes=collocation_nodes,
        quadrature_weights=quadrature_weights,
        differentiation_matrix=diff_matrix,
        barycentric_weights_for_state_nodes=bary_weights_state_nodes,
        barycentric_weights_for_collocation_nodes=bary_weights_collocation_nodes,
        lagrange_at_tau_plus_one=lagrange_at_tau_plus_one,
    )


def _evaluate_lagrange_interpolation_at_points(
    nodes: FloatArray,
    barycentric_weights: FloatArray,
    values: FloatArray,
    eval_points: float | FloatArray,
) -> FloatArray:
    is_scalar = np.isscalar(eval_points)
    eval_array = np.atleast_1d(eval_points)
    values_2d = np.atleast_2d(values)
    num_vars = values_2d.shape[0]

    result = np.zeros((num_vars, len(eval_array)), dtype=np.float64)

    for i, zeta in enumerate(eval_array):
        L_j = _evaluate_lagrange_polynomial_at_point(nodes, barycentric_weights, zeta)
        result[:, i] = np.dot(values_2d, L_j)

    return result[:, 0] if is_scalar else result



import logging


logger = logging.getLogger(__name__)


class MAPTORBaseError(Exception):
    """
    Base class for all MAPTOR-specific errors.

    All MAPTOR exceptions inherit from this class, allowing users to catch
    any MAPTOR-specific error with a single except clause.

    Args:
        message: The error message describing what went wrong
        context: Optional additional context about where the error occurred
    """

    def __init__(self, message: str, context: str | None = None) -> None:
        self.message = message
        self.context = context

        logger.debug("MAPTOR exception: %s", self._format_message())
        super().__init__(self._format_message())

    def _format_message(self) -> str:
        if self.context:
            return f"{self.message} (Context: {self.context})"
        return self.message


class ConfigurationError(MAPTORBaseError):
    """
    Raised when there is an invalid or incomplete MAPTOR configuration.

    This exception indicates that the user has provided invalid parameters,
    missing required configuration, or conflicting settings that prevent
    the problem from being solved.

    Examples:
        - Invalid polynomial degrees
        - Missing dynamics or objective function
        - Incompatible constraint specifications
        - Invalid mesh configuration
    """

    pass


class DataIntegrityError(MAPTORBaseError):
    """
    Raised when internal data corruption or inconsistency is detected.

    This exception indicates an internal MAPTOR error where data structures
    have become corrupted or inconsistent. This typically represents a bug in
    MAPTOR rather than user error.

    Examples:
        - NaN or infinite values in computed results
        - Mismatched array dimensions in internal calculations
        - Corrupted solution data structures
    """

    pass


class SolutionExtractionError(MAPTORBaseError):
    """
    Raised when solution data cannot be extracted from the optimization result.

    This exception occurs when MAPTOR fails to process the raw solver output
    into the user-friendly Solution format, typically due to unexpected solver
    behavior or corrupted optimization results.
    """

    pass


class InterpolationError(MAPTORBaseError):
    """
    Raised when adaptive mesh interpolation operations fail.

    This exception occurs during adaptive mesh refinement when MAPTOR
    cannot successfully interpolate solution data between different mesh
    configurations, typically due to numerical issues or corrupted trajectory data.
    """

    pass
